import argparse
import pandas as pd
import numpy as np
from sklearn.utils import shuffle
from sklearn.model_selection import KFold
import os

def read_data(filename, has_header=True, has_index=False, sep=',', decimal='.', missing_value='null'):
    """
    Assume file has proper format:

    Users should specify missing value if present.
    No delimiter allowed in entry.
    First line is header if has_header=True.
    First column is index if has_index=True.
    Last column is target.

    """
    data = []
    types = []
    with open(filename, 'r', encoding='unicode_escape') as f:
        lines = f.readlines()
        lines = [line.strip() for line in lines if line.strip()]  # strip empty lines
        if sep == 't':
            sep = '\t'
        first_line = lines[0].split(sep)
        first_line = list(filter(None, first_line))

        num_col = len(first_line)
        if has_header:
            header = first_line
        else:
            header = list(map(str, range(not (has_index), num_col)))
            header = ['feature_' + e for e in header] + ['target']

        if has_index:
            header = header[1:]

        for line in lines[has_header:]:
            line = line.split(sep)[has_index:]
            line = list(filter(None, line))
            line = [e.strip() for e in line]
            if len(line) < len(header): continue
            if len(types) : break
            if missing_value not in line:
                for element in line:
                    if decimal in element:
                        types.append('float')
                    elif element.isdigit():
                        types.append('int')
                    else:
                        types.append('O')

        for line in lines[has_header:]:

            if decimal == ',':
                line = line.replace(',', '.')

            line = line.split(sep)[has_index:]
            line = list(filter(None, line))
            line = [e.strip() for e in line]

            if missing_value in line or len(line) < len(header):
                continue  # skip lines contain missing values

            for i, e in enumerate(line):
                if types[i] == 'int' and '.' in e:
                    types[i] = 'float'

            data.append(tuple(line))

    dtype = list(zip(header, types))
    data = np.array(data, dtype=dtype)

    return pd.DataFrame(data)


def encode_feature(df, filename, num_bin='all'):
    """
    transform features to binary features
    all categorical values will be encoded to binary, user should mannually determine cateogorical columns

    data: pandas dataframe
    num_bin: # threshold used to encode continuous feature that has threshold over 10

    """
    root_ext = os.path.splitext(filename)
    # encode categorical features
    df_categorical = df.select_dtypes(include=object)
    if df_categorical.shape[1]:
        df_categorical = pd.get_dummies(df_categorical, drop_first=True)
    # encode continuous features
    df_real = pd.DataFrame()
    if num_bin == 'all':
        for col in df.iloc[:, :-1].select_dtypes(exclude=object).columns:
            df_real[col] = df[col]
    else:
        with open(root_ext[0] + '_intervals.txt', 'w') as f:
            for col in df.iloc[:, :-1].select_dtypes(exclude=object).columns:
                if df[col].nunique() > 10:
                    df_real[col], interval = pd.cut(df[col], bins=int(num_bin), labels=False, retbins=True)
                    f.write(col + ":  ")
                    f.write(",".join(map(str,interval)) + '\n')
                else:
                    df_real[col] = df[col]

    df_real = pd.get_dummies(df_real.astype(str), drop_first=True)
    df_binary = pd.concat([df_categorical, df_real, df.iloc[:, -1]], axis=1)

    df_binary.to_csv(root_ext[0] + '_binary.csv', index=False)

    return df_binary



def train_test_split(filename, fold=5):
    """
    split dataset to training and testing

    filename: dataset should be binarized
    """
    df = pd.read_csv(filename)
    root_ext = os.path.splitext(filename)

    for i, e in enumerate(KFold(n_splits=fold, shuffle=True, random_state=666).split(df)):
        train_index = e[0]
        test_index = e[1]
        df.iloc[train_index].to_csv(root_ext[0] + '_train_fold_' + str(i + 1) + '.csv', index=False)
        df.iloc[test_index].to_csv(root_ext[0] + '_test_fold_' + str(i + 1) + '.csv', index=False)


def create_dataset_for_evtree(filename):
    """
    :param filename: binary feature csv, no index column
    :return:
    """
    old = pd.read_csv(filename)
    new = pd.DataFrame()
    for col in old.columns[:-1]:
        new[col] = np.where(old[col] == 1, 'yes', 'no')

    data = pd.concat([new, old.iloc[:, -1]], axis=1)
    dir_path = os.path.dirname(filename) + '/evtree_data/'
    if not os.path.exists(dir_path):
        os.mkdir(dir_path)
    data.to_csv(dir_path + os.path.basename(filename),  index=False)


parser = argparse.ArgumentParser(add_help=False)
parser.add_argument("file_name", type=str, help="dataset file name")
parser.add_argument("--sep", type=str, default=",", help="dataset seperator")
parser.add_argument("--dec", type=str, default=".", help="dataset decimal")
parser.add_argument("--missing", type=str, default="nan", help="dataset missing value")
parser.add_argument("-h", "--header", action="store_false", help="dataset has no header")
parser.add_argument("-i", "--index", action="store_true", help="dataset has index")
parser.add_argument("-b", "--binarize", type=str, default='all', help="binarize raw dataset")
parser.add_argument("-s", "--split", type=int, help="train test split")
parser.add_argument("-e", "--evtree", help="generate evtree dataset",action="store_true")
parser.add_argument("-A", "--All", action="store_true", help="do all above work")

args = parser.parse_args()
file_name = args.file_name
sep = args.sep
dec = args.dec
missing = args.missing
has_header = args.header
has_index = args.index
bin_num = args.binarize
fold = args.split
generate_evtree_dataset = args.evtree

if args.All:
    df = read_data(file_name, has_header=has_header, has_index=has_index, sep=sep, decimal=dec, missing_value=missing)
    df_binary = encode_feature(df, file_name, num_bin=bin_num)
    file_name_binary = os.path.splitext(file_name)[0] + '_binary.csv'
    if not fold:
        fold = 5
    train_test_split(file_name_binary, fold=fold)

    create_dataset_for_evtree(file_name_binary)
    for f in range(fold):
        create_dataset_for_evtree(os.path.splitext(file_name_binary)[0] + '_train_fold_' + str(f + 1) + '.csv')
        create_dataset_for_evtree(os.path.splitext(file_name_binary)[0] + '_test_fold_' + str(f + 1) + '.csv')

elif generate_evtree_dataset:
    create_dataset_for_evtree(file_name)

elif fold:
    train_test_split(file_name, fold=fold)

else:
    df = read_data(file_name, has_header=has_header, has_index=has_index, sep=sep, decimal=dec, missing_value=missing)
    encode_feature(df, file_name, num_bin=bin_num)



import pandas as pd
import numpy as np
from osst.model.osst import OSST
from osst.model.metrics import harrell_c_index, uno_c_index, integrated_brier_score, cumulative_dynamic_auc, compute_ibs_per_sample
from sklearn.model_selection import train_test_split
from sksurv.ensemble import RandomSurvivalForest
from sksurv.datasets import get_x_y
import pathlib


dataset_path = "experiments/datasets/churn/churn.csv"

# read the dataset
# preprocess your data otherwise OSST will binarize continuous feature using all threshold values.
df = pd.read_csv(dataset_path)
X, event, y = df.iloc[:,:-2].values, df.iloc[:,-2].values.astype(int), df.iloc[:,-1].values
h = df.columns[:-2]
X = pd.DataFrame(X, columns=h)
event = pd.DataFrame(event)
y = pd.DataFrame(y)
_, y_sksurv = get_x_y(df, df.columns[-2:], 1)
print("X shape: ", X.shape)
# split train and test set
X_train, X_test, event_train, event_test, y_train, y_test, y_sksurv_train, y_sksurv_test \
      = train_test_split(X, event, y, y_sksurv, test_size=0.2, random_state=2024)

times_train = np.unique(y_train.values.reshape(-1))
times_test = np.unique(y_test.values.reshape(-1))
print("Train time thresholds range: ({:.1f}, {:.1f}),  Test time thresholds range: ({:.1f}, {:.1f})".format(\
    times_train[0], times_train[-1], times_test[0], times_test[-1]))

# compute reference lower bounds
ref_model = RandomSurvivalForest(n_estimators=100, max_depth=3, random_state=2024)
ref_model.fit(X_train, y_sksurv_train)
ref_S_hat = ref_model.predict_survival_function(X_train)
ref_estimates = np.array([f(times_train) for f in ref_S_hat])
ibs_loss_per_sample = compute_ibs_per_sample(event_train, y_train, event_train, y_train, ref_estimates, times_train)

labelsdir = pathlib.Path('/tmp/warm_lb_labels')
labelsdir.mkdir(exist_ok=True, parents=True)

labelpath = labelsdir / 'warm_label.tmp'
labelpath = str(labelpath)

pd.DataFrame(ibs_loss_per_sample, columns=['class_labels']).to_csv(labelpath, header='class_labels', index=None)

# fit model

config = {
    "look_ahead": True,
    "diagnostics": True,
    "verbose": False,

    "regularization": 0.01,
    "uncertainty_tolerance": 0.0,
    "upperbound": 0.0,
    "depth_budget": 5,
    "minimum_captured_points": 7,

    "model_limit": 100,
    
    "warm_LB": True,
    "path_to_labels": labelpath,
  }


model = OSST(config)
model.fit(X_train, event_train, y_train)
print("evaluate the model, extracting tree and scores", flush=True)

# evaluation
n_leaves = model.leaves()
n_nodes = model.nodes()
time = model.time
print("Model training time: {}".format(time))
print("# of leaves: {}".format(n_leaves))

print("Train IBS score: {:.6f} , Test IBS score: {:.6f}".format(\
    model.score(X_train, event_train, y_train), model.score(X_test, event_test, y_test)))

S_hat_train = model.predict_survival_function(X_train)
estimates_train = np.array([f(times_train) for f in S_hat_train])

S_hat_test = model.predict_survival_function(X_test)
estimates_test = np.array([f(times_test) for f in S_hat_test])

print("Train Harrell's c-index: {:.6f}, Test Harrell's c-index: {:.6f}".format(\
    harrell_c_index(event_train, y_train, estimates_train, times_train)[0], \
    harrell_c_index(event_test, y_test, estimates_test, times_test)[0]))

print("Train Uno's c-index: {:.6f}, Test Uno's c-index: {:.6f}".format(\
    uno_c_index(event_train, y_train, event_train, y_train, estimates_train, times_train)[0],\
    uno_c_index(event_train, y_train, event_test, y_test, estimates_test, times_test)[0]))

print("Train AUC: {:.6f}, Test AUC: {:.6f}".format(\
    cumulative_dynamic_auc(event_train, y_train, event_train, y_train, estimates_train, times_train)[0],\
    cumulative_dynamic_auc(event_train, y_train, event_test, y_test, estimates_test, times_test)[0]))

print(model.tree)


import pathlib
import platform
import distro
import subprocess
import sys
import os
import shutil


def setup(args) -> None:
    """
    Interface of the cross-platform scikit-build script
    :param args: A list of arguments passed to `setup.py`
    """
    subprocess.run([sys.executable, "setup.py"] + args).check_returncode()


def delocate_wheel(args) -> None:
    """
    Interface of `delocate-wheel` that repairs wheel files on macOS
    :param args: A list of arguments passed to `delocate-wheel`
    """
    subprocess.run(["delocate-wheel"] + args).check_returncode()


def auditwheel(args) -> None:
    """
    Interface of `auditwheel` that repairs wheel files on Ubuntu
    :param args: A list of arguments passed to "auditwheel"
    """
    subprocess.run(["auditwheel"] + args).check_returncode()


def delvewheel(args) -> None:
    """
    Interface of `delvewheel` that repairs wheel files on Windows
    :param args: A list of arguments passed to `delvewheel`
    """
    subprocess.run([sys.executable, "-m", "delvewheel"] + args).check_returncode()


def repair_wheel(wheel) -> None:
    """
    Repair the generated wheel file by copying all required dynamic libraries to it
    :param wheel: Path to the generated wheel file
    """
    # Fetch the current operating system
    system = platform.system()
    # Repair the wheel
    if system == "Darwin":
        delocate_wheel(["-w", "dist", "-v", wheel])
    elif system == "Linux":
        distribution = distro.id()
        if distribution == "ubuntu":
            auditwheel(["repair", "-w", "dist", "--plat", "linux_x86_64", wheel])
        elif distribution == "centos":
            auditwheel(["repair", "-w", "dist", "--plat", "manylinux_2_17_x86_64", wheel])
            # Remove the original wheel
            # The fixed wheel has a difference file name
            os.remove(wheel)
        else:
            print("Linux distribution {} is not supported by this script.".format(distribution))
            raise EnvironmentError
    elif system == "Windows":
        search_path = str(pathlib.Path(os.getenv("VCPKG_INSTALLATION_ROOT")) / "installed\\x64-windows\\bin")
        delvewheel(["repair", "--no-mangle-all", "--add-path", search_path, wheel, "-w", "dist"])
    else:
        print("{} is not supported.".format(system))
        raise EnvironmentError


def remove_dir_if_exists(str) -> None:
    if os.path.exists(str):
        shutil.rmtree(str)


if __name__ == '__main__':
    try:
        print(">> Cleaning the garbage...")
        remove_dir_if_exists("dist")
        remove_dir_if_exists("osst.egg-info")
        setup(["clean"])

        print(">> Rebuilding the project from scratch...")
        # `--py-limited-api=cp37` is needed otherwise installing the wheel file produced by the CentOS 7 docker
        # on other Linux distributions, such as Ubuntu, leads to an error `Not a supported wheel on this platform`.
        # On Windows, the value must match the version of the Python that builds the wheel file,
        # otherwise, wheel files will have the same name, despite being generated by different Python installations.
        api_version = "cp{}{}".format(sys.version_info.major, sys.version_info.minor)
        print("Using the CPython API {}.".format(api_version))
        setup(["bdist_wheel", "--py-limited-api={}".format(api_version), "--build-type=Release",
               "-G", "Ninja", "--", "--", "-j{}".format(os.cpu_count())])

        print(">> Adding required dynamic libraries to the wheel file...")
        wheels = os.listdir("dist")
        assert len(wheels) == 1, "The number of generated wheels is not 1. All wheels: {}.".format(wheels)
        wheel = "dist/{}".format(wheels[0])
        print("Wheel file to be fixed: {}.".format(wheel))
        repair_wheel(wheel)

        print("All done.")
        exit(0)
    except (EnvironmentError, subprocess.CalledProcessError):
        exit(1)



import pandas as pd
import numpy as np
from json import dumps, JSONEncoder
from numpy import array
from .utils import BaseEstimator
from .metrics import _integrated_brier_score
# Supporting Override for Converting Numpy Types into Python Values
class NumpyEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyEncoder, self).default(obj)


class TreeSurvivalRegressor:
    """
    Unified representation of a survival tree regressor in Python

    This class accepts a dictionary representation of a tree regressor and decodes it into an interactive object

    Additional support for encoding/decoding layer can be layers if the feature-space of the model differs from the feature space of the original data
    """

    def __init__(self, source, encoder=None, X=None, event=None, y=None):
        self.source = source  # The regressor stored in a recursive dictionary structure
        self.encoder = encoder  # Optional encoder / decoder unit to run before / after prediction

        
        if X is not None and event is not None and y is not None:  # Original training features and labels to fill in missing training loss values

            if len(event.shape) > 1:
                event = event.values.reshape(-1)
            if len(y.shape) > 1:
                y = y.values.reshape(-1)
            self.__initialize_training_loss__(X, event, y)
            # self.G = BaseEstimator(event=event, y=y, reverse=True)

    def __initialize_training_loss__(self, X, event, y):
        """
        Compares every prediction y_hat against the labels y, then incorporates the misprediction into the stored loss values
        This is used when parsing models from an algorithm that doesn't provide the training loss in the out put

        Parameters
        ---
        X : matrix-like, shape = [n_samples, m_features]
            matrix containing the training samples and features
        event: array-like, shape = [n_samples, ]
            an n-by-1 column of event associated with each sample
        y : array-like, shape = [n_samplee, ]
            column containing the correct label for each sample in X

        """
        self.G = BaseEstimator(event=event, y=y, reverse=True)
        (n, m) = X.shape
        predicted_leaf = self.predict(X)
        for node in self.__all_leaves__():
            node["loss"] = 0.0
            captured = (predicted_leaf == node['prediction'])
            # captured_X = X[captured]
            captured_event = event[captured]
            captured_y = y[captured]
            node["estimator"] = BaseEstimator(event=captured_event, y=captured_y)
            


        return

    def __find_leaf__(self, sample):
        """
        Returns
        ---
        the leaf by which this sample would be classified        
        """
        nodes = [self.source]
        while len(nodes) > 0:
            node = nodes.pop()
            if "prediction" in node:
                return node
            else:
                value = sample[node["feature"]]
                reference = node["reference"]
                if node["relation"] == "==":
                    if value == reference:
                        nodes.append(node["true"])
                    else:
                        nodes.append(node["false"])
                elif node["relation"] == ">=":
                    if value >= reference:
                        nodes.append(node["true"])
                    else:
                        nodes.append(node["false"])
                elif node["relation"] == "<=":
                    if value <= reference:
                        nodes.append(node["true"])
                    else:
                        nodes.append(node["false"])
                elif node["relation"] == ">":
                    if value > reference:
                        nodes.append(node["true"])
                    else:
                        nodes.append(node["false"])
                elif node["relation"] == "<":
                    if value < reference:
                        nodes.append(node["true"])
                    else:
                        nodes.append(node["false"])
                else:
                    raise "Unsupported relational operator {}".format(node["relation"])

    def __all_leaves__(self):
        """
        Returns
        ---
        list : a list of all leaves in this model
        """
        nodes = [self.source]
        leaf_list = []
        while len(nodes) > 0:
            node = nodes.pop()
            if "prediction" in node:
                leaf_list.append(node)
            else:
                nodes.append(node["true"])
                nodes.append(node["false"])
        return leaf_list

    def loss(self):
        """
        Returns
        ---
        real number : values between [0,1]
            the training loss of this model
        """
        return sum(node["loss"] for node in self.__all_leaves__())

    def classify(self, sample):
        """
        Parameters
        ---
        sample : array-like, shape = [m_features]
            a 1-by-m row representing each feature of a single sample

        Returns
        ---
        string : the prediction for a given sample and conditional probability (given the observations along the decision path) of it being correct
        """
        node = self.__find_leaf__(sample)
        return node["prediction"]

    def predict(self, X):
        """
        Requires
        ---
        the set of features used should be pre-encoding if an encoder is used

        Parameters
        ---
        X : matrix-like, shape = [n_samples by m_features]
            a matrix where each row is a sample to be predicted and each column is a feature to be used for prediction

        Returns
        ---
        array-like, shape = [n_sampels by 1] : a column where each element is the prediction associated with each row
        """
        if self.encoder is not None:  # Perform an encoding if an encoding unit is specified
            X = pd.DataFrame(self.encoder.encode(X.values[:, :]), columns=self.encoder.headers)

        predictions = []
        (n, m) = X.shape
        for i in range(n):
            prediction = self.classify(X.values[i, :])
            predictions.append(prediction)
        return array(predictions)

    def predict_survival_function(self, X):
        """
        Parameters
        ---
        X : matrix-like, shape = [n_samples by m_features]
            a matrix where each row is a sample to be predicted and each column is a feature to be used for prediction
        Returns
        ---
        array-like, shape = [n_sampels by 1] : a column where each element is the predicted survival function associated with each row
        """
        if self.encoder is not None:  # Perform an encoding if an encoding unit is specified
            X = pd.DataFrame(self.encoder.encode(X.values[:, :]), columns=self.encoder.headers)

        predictions = []
        (n, m) = X.shape
        for i in range(n):
            leaf_node = self.__find_leaf__(X.values[i, :])
            predictions.append(leaf_node["estimator"].predict_survival_prob)
        return array(predictions)

    def predict_cumulative_harzard_function(self, X):
        """
        Parameters
        ---
        X : matrix-like, shape = [n_samples by m_features]
            a matrix where each row is a sample to be predicted and each column is a feature to be used for prediction
        Returns
        ---
        array-like, shape = [n_sampels by 1] : a column where each element is the predicted cumulative harzard function associated with each row
        """
        if self.encoder is not None:  # Perform an encoding if an encoding unit is specified
            X = pd.DataFrame(self.encoder.encode(X.values[:, :]), columns=self.encoder.headers)

        predictions = []
        (n, m) = X.shape
        for i in range(n):
            leaf_node = self.__find_leaf__(X.values[i, :])
            predictions.append(leaf_node["estimator"].predict_cumu_harzard_prob)
        return array(predictions)
    

    def score(self, X, event, y):
        """
        Parameters
        --- 
        X : matrix-like, shape = [n_samples by m_features]
            an n-by-m matrix of sample and their features
        event: array-like, shape = [n_samples by 1] or [n_samples,]
            an n-by-1 column of event associated with each sample
        y : array-like, shape = [n_samples by 1] or [n_samples,]
            an n-by-1 column of labels associated with each sample
        Returns
        ---
        real number : the accuracy produced by applying this model overthe given dataset, with optionals for weighted accuracy
        """
        survival_functions = self.predict_survival_function(X)

        if len(event.shape) > 1:
            event = event.values.reshape(-1)
        if len(y.shape) > 1:
            y = y.values.reshape(-1)

        times = np.unique(y) 
        estimates = np.array([f(times) for f in survival_functions])
    
        return _integrated_brier_score(self.G, event, y, estimates, times)

    def __len__(self):
        """
        Returns
        ---
        natural number : The number of terminal nodes present in this tree
        """
        return self.leaves()

    def leaves(self):
        """
        Returns
        ---
        natural number : The number of terminal nodes present in this tree
        """
        leaves_counter = 0
        nodes = [self.source]
        while len(nodes) > 0:
            node = nodes.pop()
            if "prediction" in node:
                leaves_counter += 1
            else:
                nodes.append(node["true"])
                nodes.append(node["false"])
        return leaves_counter

    def nodes(self):
        """
        Returns
        ---
        natural number : The number of nodes present in this tree
        """
        nodes_counter = 0
        nodes = [self.source]
        while len(nodes) > 0:
            node = nodes.pop()
            if "prediction" in node:
                nodes_counter += 1
            else:
                nodes_counter += 1
                nodes.append(node["true"])
                nodes.append(node["false"])
        return nodes_counter

    def features(self):
        """
        Returns
        ---
        set : A set of strings each describing the features used by this model
        """
        feature_set = set()
        nodes = [self.source]
        while len(nodes) > 0:
            node = nodes.pop()
            if "prediction" in node:
                continue
            else:
                feature_set.add(node["name"])
                nodes.append(node["true"])
                nodes.append(node["false"])
        return feature_set

    def encoded_features(self):
        """
        Returns
        ---
        natural number : The number of encoded features used by the supplied encoder to represent the data set
        """
        return len(self.encoder.headers) if self.encoder is not None else None

    def maximum_depth(self, node=None):
        """
        Returns
        ---
        natural number : the length of the longest decision path in this tree. A single-node tree will return 1.
        """
        if node is None:
            node = self.source
        if "prediction" in node:
            return 1
        else:
            return 1 + max(self.maximum_depth(node["true"]), self.maximum_depth(node["false"]))

    def __str__(self):
        """
        Returns
        ---
        string : pseuodocode representing the logic of this regressor
        """
        cases = []
        for group in self.__groups__():
            predicates = []
            for name in sorted(group["rules"].keys()):
                domain = group["rules"][name]
                if domain["type"] == "Categorical":
                    if len(domain["positive"]) > 0:
                        predicates.append("{} = {}".format(name, list(domain["positive"])[0]))
                    elif len(domain["negative"]) > 0:
                        if len(domain["negative"]) > 1:
                            predicates.append(
                                "{} not in {{ {} }}".format(name, ", ".join([str(v) for v in domain["negative"]])))
                        else:
                            predicates.append("{} != {}".format(name, str(list(domain["negative"])[0])))
                    else:
                        raise "Invalid Rule"
                elif domain["type"] == "Numerical":
                    predicate = name
                    if domain["min"] != -float("INF"):
                        predicate = "{} <= ".format(domain["min"]) + predicate
                    if domain["max"] != float("INF"):
                        predicate = predicate + " < {}".format(domain["max"])
                    predicates.append(predicate)

            if len(predicates) == 0:
                condition = "if true then:"
            else:
                condition = "if {} then:".format(" and ".join(predicates))
            outcomes = []
            # for outcome, probability in group["distribution"].items():
            outcomes.append("    predicted {}: {}".format(group["name"], group["prediction"]))
            outcomes.append("    normalized loss penalty: {}".format(round(group["loss"], 3)))
            outcomes.append("    complexity penalty: {}".format(round(group["complexity"], 3)))
            result = "\n".join(outcomes)
            cases.append("{}\n{}".format(condition, result))
        return "\n\nelse ".join(cases)

    def __repr__(self):
        """
        Returns
        ---
        dictionary : The recursive dictionary used to represent the model
        """
        return self.source

    def latex(self, node=None):
        """
        Note
        ---
        This method doesn't work well for label headers that contain underscores due to underscore being a reserved character in LaTeX

        Returns
        ---
        string : A LaTeX string representing the model
        """
        if node is None:
            node = self.source
        if "prediction" in node:
            if "name" in node:
                name = node["name"]
            else:
                name = "feature_{}".format(node["feature"])
            return "[ ${}$ [ ${}$ ] ]".format(name, node["prediction"])
        else:
            if "name" in node:
                if "=" in node["name"]:
                    name = "{}".format(node["name"])
                else:
                    name = "{} {} {}".format(node["name"], node["relation"], node["reference"])
            else:
                name = "feature_{} {} {}".format(node["feature"], node["relation"], node["reference"])
            return "[ ${}$ {} {} ]".format(name, self.latex(node["true"]), self.latex(node["false"])).replace("==",
                                                                                                              " \eq ").replace(
                ">=", " \ge ").replace("<=", " \le ")

    def json(self):
        """
        Returns
        ---
        string : A JSON string representing the model
        """
        return dumps(self.source, cls=NumpyEncoder)

    def __groups__(self, node=None):
        """
        Parameters
        --- 
        node : node within the tree from which to start
        Returns
        ---
        list : Object representation of each leaf for conversion to a case in an if-then-else statement
        """
        if node is None:
            node = self.source
        if "prediction" in node:
            node["rules"] = {}
            groups = [node]
            return groups
        else:
            if "name" in node:
                name = node["name"]
            else:
                name = "feature_{}".format(node["feature"])
            reference = node["reference"]
            groups = []
            for condition_result in ["true", "false"]:
                subtree = node[condition_result]
                for group in self.__groups__(subtree):

                    # For each group, add the corresponding rule
                    rules = group["rules"]
                    if not name in rules:
                        rules[name] = {}
                    rule = rules[name]
                    if node["relation"] == "==":
                        rule["type"] = "Categorical"
                        if "positive" not in rule:
                            rule["positive"] = set()
                        if "negative" not in rule:
                            rule["negative"] = set()
                        if condition_result == "true":
                            rule["positive"].add(reference)
                        elif condition_result == "false":
                            rule["negative"].add(reference)
                        else:
                            raise "OptimalSparseSurvivalTree: Malformatted source {}".format(node)
                    elif node["relation"] == ">=":
                        rule["type"] = "Numerical"
                        if "max" not in rule:
                            rule["max"] = float("INF")
                        if "min" not in rule:
                            rule["min"] = -float("INF")
                        if condition_result == "true":
                            rule["min"] = max(reference, rule["min"])
                        elif condition_result == "false":
                            rule["max"] = min(reference, rule["max"])
                        else:
                            raise "OptimalSparseSurvivalTree: Malformatted source {}".format(node)
                    else:
                        raise "Unsupported relational operator {}".format(node["relation"])

                    # Add the modified group to the group list
                    groups.append(group)
            return groups

    


# -*- coding: utf-8 -*-
"""
Created on Sat Aug  3 22:33:17 2019

@author: chudi
"""
import gmpy2
from gmpy2 import mpz
import re
import copy
import numpy as np
import sys

labels = None
lead_one = None

"""
	Python implementation of a rule
	name: string representation of the rule
	tt: mpz representation of the samples that the rule captures
	cardinality: amount of samples in a rule
	captured: number of samples the rule captures
	predict: what outcome the rule predicts (frequentist)
	correct: percentage of samples that the rule predicts correctly
"""


class rule(object):

    def __init__(self, name, num, truthtable):
        self.num = num
        self.name = name
        self.tt = truthtable
        # subtract 1 to remove leading 1
        self.cardinality = truthtable.num_digits(2) - 1
        self.captured = gmpy2.popcount(truthtable) - 1 #number of bits with value 1 in x
        # if more than half the samples are captured
        if labels and self.captured > 0:
            self.corr = gmpy2.popcount(self.tt & labels[1].tt) - 1
            if self.corr >= (self.captured / 2):
                self.predict = 1
                self.correct = self.corr / float(self.captured)
            else:
                self.predict = 0
                self.correct = (self.captured - self.corr) / float(self.captured)
        else:
            self.predict = 1
            self.correct = 0
            self.corr = 0

"""
	Python implementation of a ruleset
	nrules: number of rules
	rules: list of rules
	tt: mpz representation of samples this ruleset captures
"""


class ruleset(object):

    def __init__(self, nrules, rules):
        self.nrules = nrules
        self.rules = rules
        self.tt = mpz(pow(2, rules[0].cardinality))
        for r in rules:
            self.tt = rule_vor(self.tt, r.tt)

"""
	Python implementation of make_default
	Returns a mpz object consisting of length ones
	Note: in order to ensure you have a leading one, pass in
	a length that is 1 greater than your number of samples
"""


def make_all_ones(length):
    ones = pow(2, length) - 1 #2^length-1
    default_tt = mpz(ones)
    return default_tt

"""
	Python implementation of make_zero
	Returns a list of mpz object of 0
"""

def make_zeros(length):
    return [mpz(0)]*length

"""
	Returns a rule that captures everything not captured in the prefix
"""
def make_default(length, prefix):
    default_tt = make_all_ones(length)
    for r in prefix:
        default_tt = rule_vandnot(default_tt, r.tt)[0]
    default_rule = rule('default', default_tt)
    return default_rule

"""
	Reads in the file and creates a ruleset out of it
"""
def parse(filename):
    data = []
    with open(filename) as f:
        for line in f:
            parsed = line.split(None, 1)
            rulename = parsed[0]
            truthtable = re.sub('[\s+]', '', parsed[1])
            # to prevent clearing of leading zeroes
            truthtable = '1' + truthtable
            bitstring = mpz(truthtable, 2)
            r = rule(rulename, bitstring)
            data.append(r)
    return data

"""
	Python implementation of rules_init
"""
def rules_init(xs, ys):
    global labels, lead_one
    labels = parse(ys)
    x = parse(xs)
    lead_one = mpz(pow(2, x[0].cardinality))
    return x

"""
	Python implementation of rule_copy
"""
def rule_copy(r):
    return copy.deepcopy(r)


"""
    Python implementation of rule_vectompz
    Convert a binary vector to a mpz object
    Note: in order to ensure you have a leading one,
    add '1' in the front
"""
def rule_vectompz(vec):
    return mpz('1' + "".join([str(i) for i in vec]), 2)

"""
    Python implementation of rule_mpztovec
    Convert a mpz object to a binary vector
"""
def rule_mpztovec(tt):
    # remove the leading one
    vec = list(tt.digits(2)[1:])
    return list(map(int, vec))

"""
    Python implementation of rule_vand
    Takes in two truthtables
    Returns the and of the truthtables
    as well as the number of ones in the and
"""
def rule_vand(tt1, tt2):
    vand = tt1 & tt2
    # subtract 1 to remove leading ones
    cnt = gmpy2.popcount(vand) - 1 #the number of bits with value 1 in x
    return vand, cnt

"""
	Python implementation of rule_vor
	Takes in two truthtables
	Returns the 'or' of the truthtables
	as well as the number of ones in the 'or'
"""
def rule_vor(tt1, tt2):
    vor = tt1 | tt2
    # subtract 1 to remove leading ones
    cnt = gmpy2.popcount(vor) - 1
    return vor, cnt


"""
	Python implementation of rule_vxor
	Takes in two truthtables
	Returns the number of ones in the 'xor'
"""
def rule_vxor(tt1, tt2):
    vxor = tt1 ^ tt2
    #
    cnt = gmpy2.popcount(vxor)
    return cnt 


"""
	Python implementation of rule_vandnot
	Takes in two truthtables
	Returns the 'and' of tt1 with the 'not' of tt2
	as well as the number of ones in the 'and'
"""
def rule_vandnot(tt1, tt2):
    # or in lead_one so we don't lose our leading digits
    vand = tt1 & (~tt2 | lead_one)
    # subtract 1 to remove leading ones
    cnt = gmpy2.popcount(vand) - 1
    return vand, cnt

"""
	Python implementation of rule_isset
"""
def rule_isset(tt, ind):
    return tt.bit_test(ind)

"""
	Python implementation of rule_print
	Prints out name, number of samples captured, and percentage
	of samples correctly predicted
	If verbose is set, also prints out the rule's truthtable
"""
def rule_print(r, verbose=False):
    print("Name: {0}, Captured: {1}, Correct: {2}".format(
        r.name, r.captured, r.correct))
    if verbose:
        rule_vector_print(r)
"""
	Python implementation of rule_vector_print
"""
def rule_vector_print(r):
    print("Truthtable: {0}".format(r.tt.digits(2)))

"""
	Python implementation of ruleset_init
"""
def ruleset_init(rules):
    rs = ruleset(len(rules), rules)
    return rs

"""
	Python implementation of ruleset_add
"""
def ruleset_add(rs, rule):
    rs.rules.append(rule)
    rs.tt = rule_vor(rs.tt, rule.tt)
    return rs

"""
	Python implementation of ruleset_delete
"""
def ruleset_delete(rs, rule):
    for i in xrange(rs.nrules):
        if rs.rules[i].name == rule.name:
            rs.rules.pop(i)
    # after deleting a rule from rs, we can no longer use the rs tt
    rs.tt = None
    return rs

"""
	Python implementation of ruleset_print
"""


def ruleset_print(rs):
    for r in rs.rules:
        rule_print(r)

"""
	Count the number of ones in an mpz object
	ensuring we strip off the leading one
"""
def count_ones(tt):
    return gmpy2.popcount(tt) - 1

"""
	Common code for accuracy and check_prefix
	Goes through the rules in a ruleset and counts how many
	samples the ruleset predicts correctly
"""


def count_corr(rs):
    corr = 0
    tot = rs.rules[0].cardinality
    unseen = make_all_ones(tot + 1)
    for r in rs.rules:
        cap = rule_vand(r.tt, unseen)[0]
        corr += rule_vand(cap, labels[r.predict].tt)[1]
        unseen = rule_vandnot(unseen, r.tt)[0]
    return corr, unseen


import numpy as np
import pandas as pd
import math
import time
import pickle
from itertools import product, compress
from gmpy2 import mpz
import sklearn.tree
import sklearn.metrics

from .rule import make_all_ones, rule_vand, rule_vectompz, count_ones

def log(tic, lines, COUNT_POP, COUNT, queue, metric, R_c, tree_old, tree_new, sorted_new_tree_rules, fname):
    "log"

    the_time = str(time.time() - tic)

    the_count_pop = str(COUNT_POP)
    the_count = str(COUNT)
    the_queue_size = str(0)  # str(len(queue))
    the_metric = str(metric)
    the_Rc = str(R_c)

    the_old_tree = str(0)  # str(sorted([leaf.rules for leaf in tree_old.cache_tree.leaves]))
    the_old_tree_splitleaf = str(0)  # str(tree_old.splitleaf)
    the_old_tree_objective = str(tree_old.cache_tree.risk)
    the_old_tree_lbound = str(tree_old.lb)
    the_new_tree = str(0)  # str(list(sorted_new_tree_rules))
    the_new_tree_splitleaf = str(0)  # str(tree_new.splitleaf)

    the_new_tree_objective = str(0)  # str(tree_new.cache_tree.risk)
    the_new_tree_lbound = str(tree_new.lb)
    the_new_tree_length = str(0)  # str(len(tree_new.cache_tree.leaves))
    the_new_tree_depth = str(0)  # str(max([len(leaf.rules) for leaf in tree_new.leaves]))

    the_queue = str(0)  # str([[ leaf.rules for leaf in thetree.leaves]  for _,thetree in queue])

    line = ";".join([the_time, the_count_pop, the_count, the_queue_size, the_metric, the_Rc,
                     the_old_tree, the_old_tree_splitleaf, the_old_tree_objective, the_old_tree_lbound,
                     the_new_tree, the_new_tree_splitleaf,
                     the_new_tree_objective, the_new_tree_lbound, the_new_tree_length, the_new_tree_depth,
                     the_queue
                     ])

    with open(fname, 'a+') as f:
        f.write(line+'\n')
        
def gini_reduction(x_mpz, y_mpz, n, rule_idx, points_cap=None):
    """
    calculate the gini reduction by each feature
    return the rank of by descending
    """
    if points_cap == None:
        points_cap = make_all_ones(n + 1)

    ndata0 = count_ones(points_cap)
    _, ndata01 = rule_vand(y_mpz, points_cap)

    p0 = ndata01 / ndata0
    gini0 = 2 * p0 * (1 - p0)

    gr = []
    for i in rule_idx:
        xi = x_mpz[i]
        l1_cap, ndata1 = rule_vand(points_cap, ~xi | mpz(pow(2, n)))

        _, ndata11 = rule_vand(l1_cap, y_mpz)

        l2_cap, ndata2 = rule_vand(points_cap, xi)

        _, ndata21 = rule_vand(l2_cap, y_mpz)

        p1 = ndata11 / ndata1 if ndata1 != 0 else 0
        p2 = ndata21 / ndata2 if ndata2 != 0 else 0
        gini1 = 2 * p1 * (1 - p1)
        gini2 = 2 * p2 * (1 - p2)
        gini_red = gini0 - ndata1 / ndata0 * gini1 - ndata2 / ndata0 * gini2
        gr.append(gini_red)

    gr = np.array(gr)
    order = list(gr.argsort()[::-1])

    odr = [rule_idx[r] for r in order]

    '''
    print("gr:", gr)
    print("order:", order)
    print("odr:", odr)
    '''

    dic = dict(zip(np.array(rule_idx)+1, odr))

    return odr, dic


def get_code(tree, feature_names, target_names, spacer_base="    "):
    """Produce psuedo-code for scikit-leant DescisionTree.
        Args
        ----
        tree -- scikit-leant DescisionTree.
        feature_names -- list of feature names.
        target_names -- list of target (class) names.
        spacer_base -- used for spacing code (default: "    ").
        Notes
        -----
        based on http://stackoverflow.com/a/30104792.
        http://chrisstrelioff.ws/sandbox/2015/06/08/decision_trees_in_python_with_scikit_learn_and_pandas.html
        """
    left = tree.tree_.children_left
    right = tree.tree_.children_right
    threshold = tree.tree_.threshold
    feats = [feature_names[i] for i in tree.tree_.feature]
    value = tree.tree_.value

    def recurse(left, right, threshold, features, node, depth):
        spacer = spacer_base * depth
        if (threshold[node] != -2):
            print((spacer + "if ( " + feats[node] + " <= " + str(threshold[node]) + " ) {"))
            if left[node] != -1:
                recurse(left, right, threshold, feats, left[node], depth + 1)
            print((spacer + "}\n" + spacer + "else {"))
            if right[node] != -1:
                recurse(left, right, threshold, feats, right[node], depth + 1)
            print((spacer + "}"))
        else:
            target = value[node]
            print((spacer + "return " + str(target)))
            for i, v in zip(np.nonzero(target)[1], target[np.nonzero(target)]):
                target_name = target_names[i]
                target_count = int(v)
                print((spacer + "return " + str(target_name) + " " + str(i) + " " \
                                                                              " ( " + str(
                    target_count) + " examples )"))

    recurse(left, right, threshold, feature_names, 0, 0)
    
def cart(x, y, name, n, P, N, lamb, w, theta, MAXDEPTH):
    clf = sklearn.tree.DecisionTreeClassifier(
            max_depth=None if MAXDEPTH == float('Inf') else MAXDEPTH,
            min_samples_split=max(math.ceil(lamb * 2 * n), 2),
            min_samples_leaf=math.ceil(lamb * n),
            max_leaf_nodes=math.floor(1 / (2 * lamb)),
            min_impurity_decrease=lamb)
    clf = clf.fit(x, y)
    nleaves_CART = (clf.tree_.node_count + 1) / 2
    pred_CART = clf.predict(x)
    if name == 'acc':
        trainout_CART = sklearn.metrics.accuracy_score(y, pred_CART)
    elif name == "bacc":
        trainout_CART = sklearn.metrics.balanced_accuracy_score(y, pred_CART)
    elif name == 'wacc':
        n_fp = sum((pred_CART == 1) & (pred_CART != y))
        n_fn = sum((pred_CART == 0) & (pred_CART != y))
        trainout_CART = 1-(n_fp + w*n_fn)/(w*P+N)
    elif name == 'f1':
        trainout_CART = sklearn.metrics.f1_score(y, pred_CART)
    elif name == 'auc_convex':
        nodes = clf.apply(x)
        metric = np.empty([3, int(nleaves_CART)], dtype=float)
        for i, num in enumerate(set(nodes)):
            idx = np.where(nodes == num)
            metric[0,i] = sum(y[idx]==1)
            metric[1,i] = sum(y[idx]==0)
            metric[2,i] = sum(y[idx]==1)/len(y[idx])
        metric = metric[:,np.argsort(metric[2,])]
        metric = np.flip(metric, axis=1)
        metric = np.cumsum(metric,axis=1)
        init = np.array([[0], [0], [0]])
        metric = np.append(init, metric, axis=1)
        trainout_CART = 0.5*sum([(metric[0,i]/P+metric[0,i-1]/P)*(metric[1,i]/N-metric[1,i-1]/N) for i in range(1,int(nleaves_CART)+1)])
    
    elif name == 'partial_auc':
        nodes = clf.apply(x)
        metric = np.empty([3, int(nleaves_CART)], dtype=float)
        for i, num in enumerate(set(nodes)):
            idx = np.where(nodes == num)
            metric[0,i] = sum(y[idx]==1)
            metric[1,i] = sum(y[idx]==0)
            metric[2,i] = sum(y[idx]==1)/len(y[idx])
        metric = metric[:,np.argsort(metric[2,])]
        metric = np.flip(metric, axis=1)
        metric = np.cumsum(metric,axis=1)
        init = np.array([[0], [0], [0]])
        metric = np.append(init, metric, axis=1)
        
        tp = [0]
        fp = [0]
        i = 1
        while fp[i-1] < N*theta and i <= int(nleaves_CART)+1:
            tp.append(metric[0,i])
            fp.append(metric[1,i])
            i += 1
            
        tp[i-1] = ((tp[i-1]-tp[i-2])/(fp[i-1]-fp[i-2]))*(N*theta-fp[i-1])+tp[i-1]
        fp[i-1] = N*theta
        trainout_CART = 0.5*sum([(tp[i]/P+tp[i-1]/P)*(fp[i]/N-fp[i-1]/N) for i in range(1,len(tp))])
     
        

    R_c = 1 - trainout_CART + lamb*nleaves_CART
    d_c = clf
    C_c = 0
    return clf, nleaves_CART, trainout_CART, R_c, d_c, C_c

def get_z(x, y, n, m):
    """
    calculate z, which is for the equivalent points bound
    z is the vector defined in algorithm 5 of the CORELS paper
    z is a binary vector indicating the data with a minority lable in its equivalent set
    """
    z = pd.DataFrame([-1] * n).values
    # enumerate through theses samples
    for i in range(n):
        # if z[i,0]==-1, this sample i has not been put into its equivalent set
        if z[i, 0] == -1:
            tag1 = np.array([True] * n)
            for j in range(m):
                rule_label = x[i][j]
                # tag1 indicates which samples have exactly the same features with sample i
                tag1 = (x[:, j] == rule_label) * tag1

            y_l = y[tag1]
            pred = int(y_l.sum() / len(y_l) >= 0.5)
            # tag2 indicates the samples in a equiv set which have the minority label
            tag2 = (y_l != pred)
            z[tag1, 0] = tag2

    z_mpz = rule_vectompz(z.reshape(1, -1)[0])
    return z_mpz



import numpy as np
import heapq
import math
import time
import copy
import sklearn.tree
import sklearn.metrics
from itertools import product, compress
from gmpy2 import mpz
from matplotlib import pyplot as plt
import pickle

from .rule import make_all_ones, rule_vand, rule_vectompz, count_ones, rule_mpztovec
from .osdt_sup import log, gini_reduction, get_code, cart, get_z

class Objective: 
    def __init__(self, name, P, N, lamb): 
        self.name = name
        self.P = P
        self.N = N
        self.lamb=lamb
    
    def leaf_predict(self, p, n, w=None):
        predict = 1
        if self.name == 'acc':
            if p<n:
                predict = 0
        elif self.name == 'bacc':
            if p/self.P <= n/self.N:
                predict = 0
        elif self.name == 'wacc':
            if w*p <= n:
                predict = 0
        elif self.name == 'f1':
            if w*p <= n:
                predict = 0      
        return predict
    
    def loss(self, FP, FN, w=None, leaves=None, theta=None):
        if self.name == 'acc':
            loss = self.acc_loss(FP, FN)
        elif self.name == 'bacc':
            loss = self.bacc_loss(FP, FN)
        elif self.name == 'wacc':
            loss = self.wacc_loss(FP, FN, w)
        elif self.name == 'f1':
            loss = self.f1_loss(FP, FN)
        elif self.name == 'auc':
            _, loss = self.ach_loss(leaves)
        elif self.name == 'pauc':
            _, loss = self.pach_loss(leaves, theta)
        return loss
    
    def risk(self, leaves, w=None, theta=None):
        FP, FN = get_false(leaves)
        risk = self.loss(FP, FN, w, leaves, theta) +  self.lamb*len(leaves)
        return risk
    
    def lower_bound(self, leaves, splitleaf, w=None, theta=None):
        FP, FN = get_fixed_false(leaves, splitleaf)
        if self.name == 'acc':
            loss = self.acc_loss(FP, FN)
        elif self.name == 'bacc':
            loss = self.bacc_loss(FP, FN)
        elif self.name == 'wacc':
            loss = self.wacc_loss(FP, FN, w)
        elif self.name == 'f1':
            loss = self.f1_loss(FP, FN)
        elif self.name == 'auc':
            ordered_fixed = order_fixed_leaves(leaves, splitleaf)
            if len(ordered_fixed) == 0:
                loss = 0
            else:
                Ps, Ns = get_split(leaves, splitleaf)
                loss = self.ach_lb(ordered_fixed, Ps, Ns)
        elif self.name == 'pauc':
            ordered_fixed = order_fixed_leaves(leaves, splitleaf)
            if len(ordered_fixed) == 0:
                loss = 1-theta
            else:
                Ps, Ns = get_split(leaves, splitleaf)
                loss = self.pach_lb(ordered_fixed, theta, Ps, Ns)
        lb = loss + self.lamb*len(leaves)
        return lb
     
    
    def incre_accu_bound(self, removed_l, new_l1, new_l2, unchanged_leaves=None, 
                         removed_leaves=None, w=None, FPu=None, FNu=None, theta=None, ld=None):
        if self.name == 'acc':
            a = self.acc_loss(removed_l.fp-new_l1.fp-new_l2.fp, 
                              removed_l.fn-new_l1.fn-new_l2.fn)
        elif self.name == 'bacc':
            a = self.bacc_loss(removed_l.fp-new_l1.fp-new_l2.fp, 
                              removed_l.fn-new_l1.fn-new_l2.fn)
        elif self.name == 'wacc':
            a = self.wacc_loss(removed_l.fp-new_l1.fp-new_l2.fp, 
                               removed_l.fn-new_l1.fn-new_l2.fn, w)
        elif self.name == 'f1':
            a = self.f1_loss(FPu+removed_l.fp, FNu+removed_l.fn) - \
                self.f1_loss(FPu+new_l1.fp+new_l2.fp, FNu+new_l1.fn+new_l2.fn)
        elif self.name == 'auc':
            removed = removed_leaves.copy()
            removed.remove(removed_l)
            _, ld_new = self.ach_loss(unchanged_leaves+removed+\
                                      [new_l1]+[new_l2])
            a = ld-ld_new
        elif self.name == 'pauc':
            removed = removed_leaves.copy()
            removed.remove(removed_l)
            _, ld_new = self.pach_loss(unchanged_leaves+removed+\
                                       [new_l1]+[new_l2], theta)
            a = ld-ld_new
        return a
    
    def leaf_support_bound(self, leaf, w=None, theta=None, 
                           FP=None, FN=None, ld=None, ordered_leaves=None):
        if self.name == 'acc':
            tau = self.acc_loss(leaf.fp, leaf.fn)
        elif self.name == 'bacc':
            tau = self.bacc_loss(leaf.fp, leaf.fn)
        elif self.name == 'wacc':
            tau = self.wacc_loss(leaf.fp, leaf.fn, w)
        elif self.name == 'f1':
            tau = ld - self.f1_loss(FP-leaf.fp, FN-leaf.fn)
        elif self.name == 'auc':
            if leaf.p == 0 or leaf.n == 0:
                tau = -1
            else:
                leaves = ordered_leaves.copy()
                leaves.remove(leaf)
                ld_new = self.ach_lb(leaves, leaf.p, leaf.n)
                tau = ld-ld_new
        elif self.name == 'pauc':
            if leaf.p == 0 or leaf.n == 0:
                tau = -1
                ld_new = ld
            else:
                leaves = ordered_leaves.copy()
                leaves.remove(leaf)
                ld_new = self.pach_lb(leaves, theta, leaf.p, leaf.n)
                tau = ld-ld_new
        #print('leaf split feature', leaf.rules)
        #print('tau:',round(tau,4), 'ld:', round(ld,4), 'ld_new:', round(ld_new, 4))
        return tau
    
    def acc_loss(self, FP, FN):
        return (FP+FN)/(self.P+self.N)
    
    def bacc_loss(self, FP, FN):
        return 0.5*(FN/self.P + FP/self.N)
    
    def wacc_loss(self, FP, FN, w):
        return (FP+w*FN)/(w*self.P+self.N)
    
    def f1_loss(self, FP, FN):
        return (FP+FN)/(2*self.P+FP-FN)
    
    def ach_loss(self, leaves):
        ordered_leaves = order_leaves(leaves)
        tp = fp = np.array([0])
        if len(leaves) > 1:
            for i in range(0, len(leaves)):
                tp = np.append(tp, tp[i]+ordered_leaves[i].p)
                fp = np.append(fp, fp[i]+ordered_leaves[i].n)
        else:
            tp = np.append(tp, self.P)
            fp = np.append(fp, self.N)
    
        loss = 1-0.5*sum([(tp[i]+tp[i-1])*(fp[i]-fp[i-1])/(self.P*self.N) \
                          for i in range(1,len(tp))])
        return ordered_leaves, loss
    
    def pach_loss(self, leaves, theta):
        ordered_leaves = order_leaves(leaves)
        tp = fp = np.array([0], dtype=float)
        if len(leaves) > 1:
            i = 0
            while fp[i] < self.N*theta and i < len(leaves):
                tp = np.append(tp, tp[i]+ordered_leaves[i].p)
                fp = np.append(fp, fp[i]+ordered_leaves[i].n)
                i += 1
            tp[i] = ((tp[i]-tp[i-1])/(fp[i]-fp[i-1]))*(self.N*theta-fp[i])+tp[i]
            fp[i] = self.N*theta
        else:
            tp = np.append(tp, self.P*theta)
            fp = np.append(fp, self.N*theta)
        loss = 1-0.5*sum([(tp[i]+tp[i-1])*(fp[i]-fp[i-1])/(self.P*self.N) \
                          for i in range(1,len(tp))])
        return ordered_leaves, loss
    
    def ach_lb(self, leaves, Ps, Ns):
        tp = np.array([0, Ps], dtype=float)
        fp = np.array([0,0], dtype=float)
        for i in range(len(leaves)):
            tp = np.append(tp, tp[i+1]+leaves[i].p)
            fp = np.append(fp, fp[i+1]+leaves[i].n)
        tp = np.append(tp, tp[len(leaves)+1]+0)
        fp = np.append(fp, fp[len(leaves)+1]+Ns)
        loss = 1- 0.5*sum([(tp[i]+tp[i-1])*(fp[i]-fp[i-1])/(self.P*self.N) for i in range(1,len(tp))])
        
        return loss
    
    def pach_lb(self, leaves, theta, Ps, Ns):
        tp = np.array([0, Ps], dtype=float)
        fp = np.array([0, 0], dtype=float)
        i = 0
        while fp[i+1] < self.N*theta:
            if i < len(leaves):
                tp = np.append(tp, tp[i+1]+leaves[i].p)
                fp = np.append(fp, fp[i+1]+leaves[i].n)
                i += 1
            else:
                tp = np.append(tp, tp[i+1]+0)
                fp = np.append(fp, fp[i+1]+Ns)
                break                         
        tp[len(tp)-1] = ((tp[len(tp)-1]-tp[len(tp)-2])/(fp[len(fp)-1]-fp[len(fp)-2]))*\
                        (self.N*theta-fp[len(fp)-1])+tp[len(tp)-1]
        fp[len(fp)-1] = self.N*theta
        loss = 1- 0.5*sum([(tp[i]+tp[i-1])*(fp[i]-fp[i-1])/(self.P*self.N) \
               for i in range(1,len(tp))])
        return loss
      

class CacheTree:
    def __init__(self, name, P, N, lamb, leaves, w=None, theta=None):
        self.name = name
        self.P = P
        self.N = N
        self.leaves = leaves
        self.w = w
        self.theta = theta
            
        bound = Objective(name, P, N, lamb)
        self.risk = bound.risk(self.leaves, self.w, self.theta)
    
    def sorted_leaves(self):
        return tuple(sorted(leaf.rules for leaf in self.leaves))
        

class Tree:
    def __init__(self, cache_tree, n, lamb, splitleaf=None, prior_metric=None):
        
        self.cache_tree = cache_tree
        self.splitleaf = splitleaf
        leaves = cache_tree.leaves
        self.H = len(leaves)
        self.risk = cache_tree.risk
        
        bound = Objective(cache_tree.name, cache_tree.P, cache_tree.N, lamb)
        self.lb = bound.lower_bound(leaves, splitleaf, cache_tree.w, cache_tree.theta)
            
        if leaves[0].num_captured == n:
            self.metric = 0                #null tree
        elif prior_metric == "objective":
            self.metric = self.risk
        elif prior_metric == "bound":
            self.metric = self.lb
        elif prior_metric == "curiosity":
            removed_leaves = list(compress(leaves, splitleaf)) #dsplit
            num_cap_rm = sum(leaf.num_captured for leaf in removed_leaves) # num captured by dsplit
            if num_cap_rm < n:
                self.metric = self.lb / ((n - num_cap_rm) / n) # supp(dun, xn)
            else:
                self.metric = self.lb / (0.01 / n) # null tree
        elif prior_metric == "entropy":
            removed_leaves = list(compress(leaves, splitleaf))
            num_cap_rm = sum(leaf.num_captured for leaf in removed_leaves)
            # entropy weighted by number of points captured
            self.entropy = [
                (-leaves[i].p * math.log2(leaves[i].p) - (1 - leaves[i].p) * math.log2(1 - leaves[i].p)) * leaves[
                    i].num_captured if leaves[i].p != 0 and leaves[i].p != 1 else 0 for i in range(self.H)]
            if num_cap_rm < n:
                self.metric = sum(self.entropy[i] for i in range(self.H) if splitleaf[i] == 0) / (
                        n - sum(leaf.num_captured for leaf in removed_leaves))
            else:
                self.metric = sum(self.entropy[i] for i in range(self.H) if splitleaf[i] == 0) / 0.01
        elif prior_metric == "gini":
            removed_leaves = list(compress(leaves, splitleaf))
            num_cap_rm = sum(leaf.num_captured for leaf in removed_leaves)
            # gini index weighted by number of points captured
            self.giniindex = [(2 * leaves[i].p * (1 - leaves[i].p))
                              * leaves[i].num_captured for i in range(self.H)]
            if num_cap_rm < n:
                self.metric = sum(self.giniindex[i] for i in range(self.H) if splitleaf[i] == 0) / (
                        n - sum(leaf.num_captured for leaf in removed_leaves))
            else:
                self.metric = sum(self.giniindex[i] for i in range(self.H) if splitleaf[i] == 0) / 0.01
        elif prior_metric == "FIFO":
            self.metric = 0
        elif prior_metric == "random":
            self.metric = np.random.uniform(0.0,1.0)
        
    def __lt__(self, other):
        # define <, which will be used in the priority queue
        return self.metric < other.metric
    

class CacheLeaf:
    def __init__(self, name, n, P, N, rules, x, y, y_mpz, z_mpz, points_cap, 
                 num_captured, lamb, support, is_feature_dead, w=None):
        self.rules = rules
        self.points_cap = points_cap
        self.num_captured = num_captured
        self.is_feature_dead = is_feature_dead
        
        _, num_ones = rule_vand(points_cap, y_mpz) #return vand and cnt
        _, num_errors = rule_vand(points_cap, z_mpz)
        '''
        print('rules:', rules)
        print("points_cap:", points_cap, "vec:", rule_mpztovec(points_cap))
        print('_:', _, "vec:", rule_mpztovec(_))
        print('num_errors',num_errors)
        '''
        self.delta = num_errors
        self.p = num_ones
        self.n = self.num_captured - num_ones
        if self.num_captured > 0 :
            self.r = num_ones/self.num_captured
        else:
            self.r = 0
        bound = Objective(name, P, N, lamb)

        if name != 'pauc':
            if num_errors > 0:
                cap = np.array(rule_mpztovec(points_cap))
                cap_i = np.where(cap == 1)[0]
                x_cap = x[cap_i]
                y_cap = y[cap_i]
                
                v = rule_mpztovec(_)
                equiv_i = np.where(np.array(v) == 1)[0]
                idx = [i for i,c in enumerate(cap_i) if c in equiv_i]
                idx = np.array(idx)
                
                unique_rows, counts = np.unique(x_cap[idx,], axis=0, return_counts=True)
                nrow = unique_rows.shape[0]
                self.equiv = np.zeros((3, nrow+2))
    
                for i in range(nrow):
                    comp = np.all(np.equal(x_cap, unique_rows[i,]), axis=1)
                    eu = np.sum(comp)
                    j = np.where(comp==True)
                    n_neg = np.sum(y_cap[j]==0)
                    n_pos = eu-n_neg
                    self.equiv[0,i] = n_pos/eu    #r = n_pos/eu
                    self.equiv[1,i] = n_pos
                    self.equiv[2,i] = n_neg
            
                self.equiv[0, nrow] = 1
                #y_i = np.where(np.array(v)==0)[0]
                #equiv_not_i = [i for i,c in enumerate(cap_i) if c not in equiv_i]
                self.equiv[1, nrow] = sum(y_cap==1) - sum(self.equiv[1,i] for i in range(nrow))
                self.equiv[2, nrow+1] = sum(y_cap==0) - sum(self.equiv[2,i] for i in range(nrow))
            else:
                self.equiv = np.zeros((3, 2))
                self.equiv[0,0] = 1
                self.equiv[1,0] = self.p
                self.equiv[2,1] = self.n
    
        if self.num_captured:
            self.pred = bound.leaf_predict(self.p, self.n, w)
            if self.pred == 0:
                self.fp = 0
                self.fn = self.p
            else:
                self.fp = self.n
                self.fn = 0
        else:
            self.pred = 0
            self.fp = 0
            self.fn = self.p
                
class EquivPoints:
    def __init__(self, r, p, n):
        self.r = r
        self.p = p
        self.n = n


def get_false(leaves):
    FP = sum([l.fp for l in leaves])
    FN = sum([l.fn for l in leaves])
    return FP, FN
    
def get_fixed_false(leaves, splitleaf):
    FPu = sum([leaves[i].fp for i in range(len(leaves)) if splitleaf[i]==0])
    FNu = sum([leaves[i].fn for i in range(len(leaves)) if splitleaf[i]==0])
    return FPu, FNu
    
def get_split(leaves, splitleaf):
    Ps = sum([leaves[i].p for i in range(len(leaves)) if splitleaf[i]==1])
    Ns = sum([leaves[i].n for i in range(len(leaves)) if splitleaf[i]==1])
    return Ps, Ns
                
def order_leaves(leaves):
    return sorted([l for l in leaves], key=lambda x:x.r, reverse=True)

def order_fixed_leaves(leaves, splitleaf):
    leaf_fixed = [not split for split in splitleaf]
    fixed_set = list(compress(leaves, leaf_fixed))
    ordered_fixed = order_leaves(fixed_set)
    return ordered_fixed

def equiv_lb(name, leaves, splitleaf, P, N, lamb, w):
    split_set = list(compress(leaves, splitleaf))
    #leaf_fixed = [not split for split in splitleaf]
    
    leaf_equiv_fp = 0
    leaf_equiv_fn = 0
    
    bound = Objective(name, P, N, lamb)
    
    for i in split_set:
        #print(rule_mpztovec(i.points_cap))
        equiv = i.equiv # equiv is the array row=3 and col=equivset+2
        if equiv.shape[1]-2 > 0:
            for j in range(equiv.shape[1]-2):
                pred = bound.leaf_predict(equiv[1,j], equiv[2,j], w)
                if pred == 0:
                    leaf_equiv_fp += 0
                    leaf_equiv_fn += equiv[1,j]
                else:
                    leaf_equiv_fp += equiv[2,j]
                    leaf_equiv_fn += 0
        #print('delta_fp:', leaf_equiv_fp, 'delta_fn:', leaf_equiv_fn)
        
    return leaf_equiv_fp, leaf_equiv_fn

    
def ach_equiv_lb(leaves, splitleaf, P, N, lamb):
    split_set = list(compress(leaves, splitleaf))
    leaf_fixed = [not split for split in splitleaf]
    fixed_set = list(compress(leaves, leaf_fixed))
    Ps1 = 0
    Ns1 = 0
    for i in split_set:
        equiv = i.equiv # equiv is the array row=3 and col=equivset+2
        if equiv.shape[1]-2 > 0:
            for j in range(equiv.shape[1]-2):
                leaf_equiv = EquivPoints(equiv[0,j], equiv[1,j], equiv[2,j])
                fixed_set.append(leaf_equiv)
        Ps1 += equiv[1, equiv.shape[1]-2]
        Ns1 += equiv[2, equiv.shape[1]-1]
    ordered_leaves = order_leaves(fixed_set)
    
    tp = np.array([0, Ps1], dtype=float)
    fp = np.array([0,0], dtype=float)
    for i in range(len(ordered_leaves)):
        tp = np.append(tp, tp[i+1]+ordered_leaves[i].p)
        fp = np.append(fp, fp[i+1]+ordered_leaves[i].n)
    tp = np.append(tp, tp[len(ordered_leaves)+1]+0)
    fp = np.append(fp, fp[len(ordered_leaves)+1]+Ns1)
    
    loss = 1- 0.5*sum([(tp[i]+tp[i-1])*(fp[i]-fp[i-1])/(P*N) for i in range(1,len(tp))])
    return loss + lamb*len(leaves)



def generate_new_splitleaf(name, P, N, unchanged_leaves, removed_leaves, new_leaves, lamb,
                           incre_support, w, theta):
    
    n_removed_leaves = len(removed_leaves)  #dsplit
    n_unchanged_leaves = len(unchanged_leaves) #dun
    n_new_leaves = len(new_leaves)
    n_new_tree_leaves = n_unchanged_leaves + n_new_leaves #H'
    splitleaf1 = [0] * n_unchanged_leaves + [1] * n_new_leaves
    
    bound = Objective(name, P, N, lamb)
    FPu = None
    FNu = None
    ld = None
    if name == 'f1':
        FPu, FNu = get_false(unchanged_leaves)
    if name == 'auc':
        _, ld = bound.ach_loss(unchanged_leaves+removed_leaves)
    if name == 'pauc':
        _, ld = bound.pach_loss(unchanged_leaves+removed_leaves, theta)
    
    sl = []
    for i in range(n_removed_leaves):
        splitleaf = [0]*n_new_tree_leaves
        removed_l = removed_leaves[i]
        new_l1 = new_leaves[2*i]
        new_l2 = new_leaves[2*i+1]
        a = bound.incre_accu_bound(removed_l, new_l1, new_l2, unchanged_leaves, 
                         removed_leaves, w, FPu, FNu, theta, ld)
        if not incre_support:
            a = float('Inf')
        if a <= lamb:
            splitleaf[n_unchanged_leaves + 2*i] = 1
            splitleaf[n_unchanged_leaves + 2*i + 1] = 1
            sl.append(splitleaf)
        else:
            sl.append(splitleaf1)
    return sl

def get_cannot_split(name, P, N, lamb, m, new_tree_leaves, MAXDEPTH, w, theta):
    bound = Objective(name, P, N, lamb)
    FP = None
    FN = None
    ld = None
    ordered_leaves = None
    if name == 'f1':
        FP, FN = get_false(new_tree_leaves)
        ld = bound.f1_loss(FP, FN)
    if name == 'auc':
        ordered_leaves, ld = bound.ach_loss(new_tree_leaves)
    if name == 'pauc':
        ordered_leaves, ld = bound.pach_loss(new_tree_leaves, theta)
        
    cannot_split = [len(l.rules) >= MAXDEPTH or bound.leaf_support_bound(l, w, theta, \
                    FP, FN, ld, ordered_leaves) < lamb or\
                    all([l.is_feature_dead[r-1] for r in range(1, m+1) if r not in \
                    map(abs, l.rules)]) for l in new_tree_leaves]
    return cannot_split



def bbound(x, y, name, lamb, prior_metric=None, w=None, theta=None, MAXDEPTH=float('Inf'), 
           MAX_NLEAVES=float('Inf'), niter=float('Inf'), logon=False,
           support=True, incre_support=True, accu_support=True, equiv_points=True,
           lookahead=True, lenbound=True, R_c0 = 1, timelimit=float('Inf'), init_cart = True,
           saveTree = False, readTree = False):

    x0 = copy.deepcopy(x)
    y0 = copy.deepcopy(y)

    tic = time.time()

    m = x.shape[1] # number of features
    n = len(y)
    P = np.count_nonzero(y)
    N = n-P

    x_mpz = [rule_vectompz(x[:, i]) for i in range(m)]
    y_mpz = rule_vectompz(y)

    # order the columns by descending gini reduction
    idx, dic = gini_reduction(x_mpz, y_mpz, n, range(m))
    #idx, dic = get_variable_importance(x, y)
    
    x = x[:, idx]
    x_mpz = [x_mpz[i] for i in idx]
    
    z_mpz = get_z(x,y,n,m)


    lines = []  # a list for log
    leaf_cache = {}  # cache leaves
    tree_cache = {}  # cache trees

    # initialize the queue to include just empty root
    queue = []
    root_leaf = CacheLeaf(name, n, P, N, (), x, y, y_mpz, z_mpz, make_all_ones(n + 1), 
                          n, lamb, support, [0] * m, w)
    d_c = CacheTree(name, P, N, lamb=lamb, leaves=[root_leaf], w=w, theta=theta)
    R_c = d_c.risk
    tree0 = Tree(cache_tree=d_c, n=n, lamb=lamb,splitleaf=[1], prior_metric=prior_metric)
    heapq.heappush(queue, (tree0.metric, tree0))
    
    best_is_cart = False  # a flag for whether or not the best is the initial CART
    if init_cart: 
        clf, nleaves_CART, trainout_CART, R_c, d_c, C_c = cart(x0, y0, name, n, P, N, lamb, w, theta, MAXDEPTH)
        time_c = time.time() - tic
        best_is_cart = True
        print('risk of cart:', R_c)
    else:
        C_c=0
        clf=None
        time_c = time.time()
        
    if readTree:
        with open('tree.pkl', 'rb') as f:
            d_c = pickle.load(f)
        R_c = d_c.risk

        with open('leaf_cache.pkl', 'rb') as f:
            leaf_cache = pickle.load(f)

        sorted_new_tree_rules = tuple(sorted(leaf.rules for leaf in d_c.leaves))
        tree_cache[sorted_new_tree_rules] = True

        tree_p = Tree(cache_tree=d_c, n=n, lamb=lamb, 
                      splitleaf=[1]*len(d_c.leaves), prior_metric=prior_metric)

        heapq.heappush(queue, (tree_p.metric, tree_p))
        '''
        print("PICKEL>>>>>>>>>>>>>", [leaf.rules for leaf in d_c.leaves])
        print('R_c:', R_c)
        print('lower_bound:', tree_p.lb)
        print('lookahead:',tree_p.lb+lamb*sum(tree_p.splitleaf))
        '''
        #print("leaf_cache:", leaf_cache)

        C_c = 0
        time_c = time.time() - tic
        
    if R_c0 < R_c:
        R_c = R_c0

    
    leaf_cache[()] = root_leaf

    COUNT = 0  # count the total number of trees in the queue
    COUNT_POP = 0 # number of tree poped from queue (# of tree checked)
    COUNT_UNIQLEAVES = 0
    COUNT_LEAFLOOKUPS = 0
    
    if logon:
        header = ['time', '#pop', '#push', 'queue_size', 'metric', 'R_c',
                  'the_old_tree', 'the_old_tree_splitleaf', 'the_old_tree_objective', 'the_old_tree_lbound',
                  'the_new_tree', 'the_new_tree_splitleaf',
                  'the_new_tree_objective', 'the_new_tree_lbound', 'the_new_tree_length', 'the_new_tree_depth', 'queue']

        fname = "_".join([name, str(m), str(n), prior_metric,
                          str(lamb), str(MAXDEPTH), str(init_cart), ".txt"])
        with open(fname, 'w') as f:
            f.write('%s\n' % ";".join(header))
    
    bound = Objective(name, P, N, lamb)
    
    #len_queue=[]
    #time_queue=[]
    #count_tree = []
    #time_realize_best_tree=[time_c]
    #R_best_tree=[R_c]
    #best_tree = [d_c]

    while queue and COUNT < niter and time.time() - tic < timelimit:
        '''
        print(len(queue))
        for metric, t in queue:
            print(metric, [l.rules for l in t.cache_tree.leaves], t.splitleaf)
        '''
        metric, tree = heapq.heappop(queue)
        

        COUNT_POP = COUNT_POP + 1
        #count_tree.append(COUNT_POP)
        
        leaves = tree.cache_tree.leaves
        leaf_split = tree.splitleaf       
        removed_leaves = list(compress(leaves, leaf_split))
        old_tree_length = len(leaf_split)
        new_tree_length = len(leaf_split) + sum(leaf_split)
        
        # prefix-specific upper bound on number of leaves
        if lenbound and new_tree_length >= min(old_tree_length + math.floor((R_c - tree.lb) / lamb),
                                               2**m):
            continue

        n_removed_leaves = sum(leaf_split)
        n_unchanged_leaves = old_tree_length - n_removed_leaves
        
        #print("num in queue:", len(queue))
        #print(time.time()-tic)
        #len_queue.append(len(queue))
        #time_queue.append(time.time()-tic)
        
        
        '''equivalent points bound + lookahead bound'''        
        lambbb = lamb if lookahead else 0
        
        if (name != 'auc') and (name != 'pauc'):
            FPu, FNu = get_fixed_false(leaves, leaf_split)
            if equiv_points:
                delta_fp, delta_fn = equiv_lb(name, leaves, leaf_split, P, N, lamb, w) 
                #print('delta_fp:', delta_fp)
                #print('delta_fn:', delta_fn)
            else:
                delta_fp=0
                delta_fn=0
            
            if (bound.loss(FPu+delta_fp, FNu+delta_fn, w)+ (old_tree_length+n_removed_leaves) * lambbb >= R_c):
                continue
  
        if (name == 'auc'): 
            if (ach_equiv_lb(leaves, leaf_split, P, N, lamb) + n_removed_leaves*lambbb >= R_c):
                continue

        
        if (name == 'pauc') and (tree.lb + n_removed_leaves * lambbb>= R_c):
            continue

        leaf_no_split = [not split for split in leaf_split]
        unchanged_leaves = list(compress(leaves, leaf_no_split))

        # Generate all assignments of rules to the leaves that are due to be split
        rules_for_leaf = [set(range(1, m + 1)) - set(map(abs, l.rules)) -
                          set([i+1 for i in range(m) if l.is_feature_dead[i] == 1]) for l in removed_leaves]

        for leaf_rules in product(*rules_for_leaf):

            if time.time() - tic >= timelimit:
                break

            new_leaves = []
            flag_increm = False  # a flag for jump out of the loops (incremental support bound)
            for rule, removed_leaf in zip(leaf_rules, removed_leaves):

                rule_index = rule - 1
                tag = removed_leaf.points_cap  # points captured by the leaf's parent leaf

                for new_rule in (-rule, rule):
                    new_rule_label = int(new_rule > 0)
                    new_rules = tuple(
                        sorted(removed_leaf.rules + (new_rule,)))
                    if new_rules not in leaf_cache:

                        COUNT_UNIQLEAVES = COUNT_UNIQLEAVES+1

                        tag_rule = x_mpz[rule_index] if new_rule_label == 1 else ~(x_mpz[rule_index]) | mpz(pow(2, n))
                        new_points_cap, new_num_captured = rule_vand(tag, tag_rule)

                        #parent_is_feature_dead =
                        new_leaf = CacheLeaf(name, n, P, N, new_rules, x, y, y_mpz, z_mpz, new_points_cap, new_num_captured,
                                             lamb, support, removed_leaf.is_feature_dead.copy(), w)
                        leaf_cache[new_rules] = new_leaf
                        new_leaves.append(new_leaf)
                    else:

                        COUNT_LEAFLOOKUPS = COUNT_LEAFLOOKUPS+1

                        new_leaf = leaf_cache[new_rules]
                        new_leaves.append(new_leaf)

                    '''
                    # Lower bound on classification accuracy
                    # if (new_leaf.num_captured) / n <= lamb:
                    # accu_support == theorem 9 in OSDT, check if feature dead, not derived yet
                    
                    if accu_support == True and (new_leaf.num_captured - new_leaf.num_captured_incorrect) / n <= lamb:

                        removed_leaf.is_feature_dead[rule_index] = 1

                        flag_increm = True
                        break
                    '''    

                if flag_increm:
                    break

            if flag_increm:
                continue

            new_tree_leaves = unchanged_leaves + new_leaves

            sorted_new_tree_rules = tuple(sorted(leaf.rules for leaf in new_tree_leaves))

            if sorted_new_tree_rules in tree_cache:
                continue
            else:
                tree_cache[sorted_new_tree_rules] = True

            child = CacheTree(name, P, N, lamb, new_tree_leaves, w=w, theta=theta)
            
            #print([l.rules for l in child.leaves])

            R = child.risk
            
            #print("R:", R, "R_c:", R_c)
            #time_realize_best_tree.append(time.time()-tic)
            #R_best_tree.append(R)
            
            
            if R < R_c:
                d_c = child
                #best_tree.append([leaf.rules for leaf in d_c.leaves])
                #R_best_tree.append(R)
                #time_realize_best_tree.append(time.time()-tic)
                R_c = R
                C_c = COUNT + 1
                time_c = time.time() - tic
                

                best_is_cart = False

            # generate the new splitleaf for the new tree
            sl = generate_new_splitleaf(name, P, N, unchanged_leaves, removed_leaves, new_leaves,
                                        lamb, incre_support, w, theta) # a_j

            cannot_split = get_cannot_split(name, P, N, lamb, m, new_tree_leaves, 
                                            MAXDEPTH, w, theta)
                

            # For each copy, we don't split leaves which are not split in its parent tree.
            # In this way, we can avoid duplications.
            can_split_leaf = [(0,)] * n_unchanged_leaves + \
                             [(0,) if cannot_split[i]
                              else (0, 1) for i in range(n_unchanged_leaves, new_tree_length)]
            # Discard the first element of leaf_splits, since we must split at least one leaf
            new_leaf_splits0 = np.array(list(product(*can_split_leaf))[1:])#sorted(product(*can_split_leaf))[1:]
            len_sl = len(sl)
            if len_sl == 1:
                # Filter out those which split at least one leaf in dp (d0)
                new_leaf_splits = [ls for ls in new_leaf_splits0
                                   if np.dot(ls, sl[0]) > 0]
            else:
                # Filter out those which split at least one leaf in dp and split at least one leaf in d0
                new_leaf_splits = [ls for ls in new_leaf_splits0
                                   if all([np.dot(ls, sl[i]) > 0 for i in range(len_sl)])]

            for new_leaf_split in new_leaf_splits:
                # construct the new tree
                tree_new = Tree(cache_tree=child, n=n, lamb=lamb,
                                splitleaf=new_leaf_split, prior_metric=prior_metric)
                '''
                print('tree_lb:', round(tree_new.lb, 4), 
                      'tree_risk:', round(tree.cache_tree.risk, 4))
                '''
                #print('tree_rules_x8:', [l.rules for l in tree.cache_tree.leaves])
                
                
                # MAX Number of leaves
                if len(new_leaf_split)+sum(new_leaf_split) > MAX_NLEAVES:
                    continue

                COUNT = COUNT + 1
                #print([l.rules for l in tree_new.cache_tree.leaves], tree_new.splitleaf)
                '''
                if (COUNT <= 22):
                    print([l.rules for l in tree_new.cache_tree.leaves], 
                          tree_new.splitleaf, round(tree_new.lb, 4), 
                          round(tree_new.cache_tree.risk,4), round(tree_new.metric, 4), 
                          round(metric,4), [l.rules for l in tree.cache_tree.leaves])
                
                if (COUNT ==22)|(COUNT == 21)|(COUNT==20):
                    for metric, t in queue:
                        print(metric, [l.rules for l in t.cache_tree.leaves], t.splitleaf)
                   
                if COUNT == 22:
                    print('123455667677')
                    return
                '''
                # heapq.heappush(queue, (2*tree_new.metric - R_c, tree_new))
                heapq.heappush(queue, (tree_new.metric, tree_new))
                
                if logon:
                    log(tic, lines, COUNT_POP, COUNT, queue, metric, R_c, tree, tree_new, sorted_new_tree_rules, fname)
                
                if COUNT % 1000000 == 0:
                    print("COUNT:", COUNT)
        #print('COUNT:', COUNT)

    totaltime = time.time() - tic

    if not best_is_cart:

        accu = 1-(R_c-lamb*len(d_c.leaves))

        leaves_c = [leaf.rules for leaf in d_c.leaves]
        pred_c = [leaf.pred for leaf in d_c.leaves]

        num_captured = [leaf.num_captured for leaf in d_c.leaves]

        #num_captured_incorrect = [leaf.num_captured_incorrect for leaf in d_c.leaves]

        nleaves = len(leaves_c)
    else:
        accu = trainout_CART
        leaves_c = 'NA'
        pred_c = 'NA'
        get_code(d_c, ['x'+str(i) for i in range(1, m+1)], [0, 1])
        num_captured = 'NA'
        #num_captured_incorrect = 'NA'
        nleaves = nleaves_CART
        
    if saveTree:
        with open('tree.pkl', 'wb') as f:
            pickle.dump(d_c, f)
        with open('leaf_cache.pkl', 'wb') as f:
            pickle.dump(leaf_cache, f)
        

    '''
    print(">>> log:", logon)
    print(">>> support bound:", support)
    print(">>> accu_support:", accu_support)
    print(">>> accurate support bound:", incre_support)
    print(">>> equiv points bound:", equiv_points)
    print(">>> lookahead bound:", lookahead)
    print("prior_metric=", prior_metric)
    '''
    print("loss function:", name)
    print("lambda: ", lamb)
    print("COUNT_UNIQLEAVES:", COUNT_UNIQLEAVES)
    print("COUNT_LEAFLOOKUPS:", COUNT_LEAFLOOKUPS)
    print("total time: ", totaltime)   
    print("leaves: ", leaves_c)
    print("num_captured: ", num_captured)
    print("prediction: ", pred_c)
    print("Objective: ", R_c)
    print(name, ": ", accu)
    print("COUNT of the best tree: ", C_c)
    print("time when the best tree is achieved: ", time_c)
    print("TOTAL COUNT: ", COUNT)

    return leaves_c, pred_c, dic, nleaves, m, n, totaltime, time_c, R_c, COUNT, C_c, \
            accu, best_is_cart, clf#, len_queue, time_queue, \
            #time_realize_best_tree, R_best_tree, count_tree, best_tree


def predict(name, leaves_c, prediction_c, nleaves, dic, x, y, best_is_cart, clf, w=None, theta=None, logon=False):
    """
    :param leaves_c:
    :param dic:
    :return:
    """
    P = np.count_nonzero(y)
    N = len(y) - P
    if best_is_cart:
        yhat = clf.predict(x)
        
        n_fp = sum((yhat == 1) & (yhat != y))
        n_fn = sum((yhat == 0) & (yhat != y))
        n_tp = sum((yhat == 1) & (yhat == y))
        
        if name == 'acc':
            out = sklearn.metrics.accuracy_score(y, yhat)
        elif name == "bacc":
            out = sklearn.metrics.balanced_accuracy_score(y, yhat)
        elif name == 'wacc':
            #print(n_fp,n_fn,P,N,w)
            out = 1-(n_fp + w*n_fn)/(w*P+N)
        elif name == 'f1':
            out = sklearn.metrics.f1_score(y, yhat)
        elif name == 'auc':
            nodes = clf.apply(x)
            metric = np.empty([3, int(nleaves)], dtype=float)
            for i, num in enumerate(set(nodes)):
                idx = np.where(nodes == num)
                metric[0,i] = sum(y[idx]==1)
                metric[1,i] = sum(y[idx]==0)
                metric[2,i] = sum(y[idx]==1)/len(y[idx])
            metric = metric[:,np.argsort(metric[2,])]
            metric = np.flip(metric, axis=1)
            metric = np.cumsum(metric,axis=1)
            init = np.array([[0], [0], [0]])
            metric = np.append(init, metric, axis=1)
            tp = metric[0, :]
            fp = metric[1, :]
            out = 0.5*sum([(metric[0,i]/P+metric[0,i-1]/P)*\
                                 (metric[1,i]/N-metric[1,i-1]/N) for i in \
                                 range(1,metric.shape[1])])
        elif name == 'pauc':
            nodes = clf.apply(x)
            metric = np.zeros([3, int(nleaves)], dtype=float)
            for i, num in enumerate(set(nodes)):
                idx = np.where(nodes == num)
                metric[0,i] = sum(y[idx]==1)
                metric[1,i] = sum(y[idx]==0)
                metric[2,i] = sum(y[idx]==1)/len(y[idx])
                #print(metric[0,i], metric[1,i], metric[2,i])
            metric = metric[:,np.argsort(metric[2,])]
            metric = np.flip(metric, axis=1)
            metric = np.cumsum(metric,axis=1)
            init = np.array([[0], [0], [0]])
            metric = np.append(init, metric, axis=1)
            '''
            tp = [0]
            fp = [0]
            i = 1
            while fp[i-1] < N*theta and i <= int(nleaves)+1:
                tp.append(metric[0,i])
                fp.append(metric[1,i])
                i += 1
            
            tp[i-1] = ((tp[i-1]-tp[i-2])/(fp[i-1]-fp[i-2]))*(N*theta-fp[i-1])+tp[i-1]
            fp[i-1] = N*theta
            out = 0.5*sum([(tp[i]+tp[i-1])*(fp[i]-fp[i-1])/(P*N) for i in range(1,len(tp))])   
            '''
            tp = metric[0, :]
            fp = metric[1, :]
            out = 0.5*sum([(metric[0,i]/P+metric[0,i-1]/P)*\
                           (metric[1,i]/N-metric[1,i-1]/N) for i in \
                           range(1,metric.shape[1])])
        
        print("Best is cart! Testing", name, round(out,4))
        print("P=", P, "N=", N, "FP=", n_fp, "FN=", n_fn)
        if (name=='auc') or (name == 'pauc'):
            print('tp=', tp)
            print('fp=',fp)
            plt.plot([f/N for f in fp], [f/P for f in tp], 'go-', linewidth=2)
        print(">>>>>>>>>>>>>>>>>>>>>>>")

        return yhat, out

    n = x.shape[0]

    caps = []

    for leaf in leaves_c:
        cap = np.array([1] * n)
        for feature in leaf:
            idx = dic[abs(feature)]
            feature_label = int(feature > 0)
            cap = (x[:, idx] == feature_label) * cap
        caps.append(cap)

    yhat = np.array([1] * n)
    cap_pos = []
    cap_neg = []

    for j in range(len(caps)):
        idx_cap = [i for i in range(n) if caps[j][i] == 1]
        yhat[idx_cap] = prediction_c[j]
        cap_pos.append(sum(y[idx_cap]))
        cap_neg.append(len(idx_cap)-sum(y[idx_cap]))

    n_fp = sum((yhat == 1) & (yhat != y))
    n_fn = sum((yhat == 0) & (yhat != y))
    #n_tp = sum((yhat == 1) & (yhat == y))
    
    
    if name == 'acc':
        out = sklearn.metrics.accuracy_score(y, yhat)
    elif name == "bacc":
        out = sklearn.metrics.balanced_accuracy_score(y, yhat)
    elif name == 'wacc':
        out = 1-(n_fp + w*n_fn)/(w*P+N)
    elif name == 'f1':
        out = sklearn.metrics.f1_score(y, yhat)
    elif name == 'auc':
        '''
        n_p = []
        n_n = []
        for i in range(len(caps)):
            idx = np.where(caps[i]==1)
            n_p.append(sum(y[idx]==1))
            n_n.append(sum(y[idx]==0))
        '''    
        metric = np.empty([3, len(caps)], dtype=float)
        for i in range(len(caps)):
            idx = np.where(caps[i]==1)
            metric[0,i] = sum(y[idx]==1)
            metric[1,i] = sum(y[idx]==0)
            metric[2,i] = sum(y[idx]==1)/len(y[idx])
        
        metric = metric[:,np.argsort(metric[2,])]
        metric = np.flip(metric, axis=1)
        metric = np.cumsum(metric,axis=1)
        init = np.array([[0], [0], [0]])
        metric = np.append(init, metric, axis=1)  
        tp = metric[0, :]
        fp = metric[1, :]
        out = 0.5*sum([(metric[0,i]+metric[0,i-1])*(metric[1,i]-metric[1,i-1])/(P*N) for i in range(1,len(tp))])
    
        
    elif name == 'pauc':
        metric = np.empty([3, len(caps)], dtype=float)
        for i in range(len(caps)):
            idx = np.where(caps[i]==1)
            metric[0,i] = sum(y[idx]==1)
            metric[1,i] = sum(y[idx]==0)
            metric[2,i] = sum(y[idx]==1)/len(y[idx])
        
        metric = metric[:,np.argsort(metric[2,])]
        metric = np.flip(metric, axis=1)
        metric = np.cumsum(metric,axis=1)
        init = np.array([[0], [0], [0]])
        metric = np.append(init, metric, axis=1)
        '''
        tp = [0]
        fp = [0]
        i = 1
        while fp[i-1] < N*theta and i <= len(caps)+1:
            tp.append(metric[0,i])
            fp.append(metric[1,i])
            i += 1          
        tp[i-1] = ((tp[i-1]-tp[i-2])/(fp[i-1]-fp[i-2]))*(N*theta-fp[i-1])+tp[i-1]
        fp[i-1] = N*theta
        out = 0.5*sum([(tp[i]/P+tp[i-1]/P)*(fp[i]/N-fp[i-1]/N) for i in range(1,len(tp))])
        '''
        tp = metric[0, :]
        fp = metric[1, :]
        out = 0.5*sum([(metric[0,i]+metric[0,i-1])*(metric[1,i]-metric[1,i-1])/(P*N) for i in range(1,len(tp))])
    
    if logon==True:
        fname = '_'.join(['pred', name, str(P+N), '.txt'])    
        with open(fname, 'w') as f:
            f.write(';'.join(["P=", str(P), "N=", str(N), 
                              "FP=", str(n_fp), "FN=", str(n_fn), 
                    name, "=", str(round(out, 4))]))
    print('Leaf:', leaves_c)
    print('captured pos:', cap_pos)
    print('captured neg:', cap_neg)        
    print("Testing", name, round(out,4))
    #print("Testing", "auc:", round(out,4))
    if name != 'auc' and name != 'pauc':
        print("P=", P, "N=", N, "FP=", n_fp, "FN=", n_fn)
        plt.plot([0, n_fp/N, 1], [0, (P-n_fn)/P, 1], 'go-', linewidth=2)
    else:
        #print("P=", P, "N=", N, "l.p=", n_p, "l.n=", n_n)
        print('P=',P, "N=", N, 'tp=', tp, "fp=", fp)
        plt.plot([f/N for f in fp], [f/P for f in tp], 'go-', linewidth=2)
    print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")

    return yhat, out


# class Objective:
#     def __init__(self, name, P, N, lamb):
#         self.name = name
#         self.P = P
#         self.N = N
#         self.lamb = lamb

#     def leaf_predict(self, p, n, w=None):
#         predict = 1
#         if self.name == "acc":
#             if p < n:
#                 predict = 0
#         elif self.name == "bacc":
#             if p / self.P <= n / self.N:
#                 predict = 0
#         elif self.name == "wacc":
#             if w * p <= n:
#                 predict = 0
#         elif self.name == "f1":
#             if w * p <= n:
#                 predict = 0
#         return predict

#     def loss(self, FP, FN, w=None, leaves=None, theta=None):
#         if self.name == "acc":
#             loss = self.acc_loss(FP, FN)
#         elif self.name == "bacc":
#             loss = self.bacc_loss(FP, FN)
#         elif self.name == "wacc":
#             loss = self.wacc_loss(FP, FN, w)
#         elif self.name == "f1":
#             loss = self.f1_loss(FP, FN)
#         elif self.name == "auc_convex":
#             _, loss = self.ach_loss(leaves)
#         elif self.name == "partial_auc":
#             _, loss = self.pach_loss(leaves, theta)
#         return loss

#     def risk(self, leaves, w=None, theta=None):
#         FP, FN = get_false(leaves)
#         risk = self.loss(FP, FN, w, leaves, theta) + self.lamb * len(leaves)
#         return risk

#     def lower_bound(self, leaves, splitleaf, w=None, theta=None):
#         FP, FN = get_fixed_false(leaves, splitleaf)
#         if self.name == "acc":
#             loss = self.acc_loss(FP, FN)
#         elif self.name == "bacc":
#             loss = self.bacc_loss(FP, FN)
#         elif self.name == "wacc":
#             loss = self.wacc_loss(FP, FN, w)
#         elif self.name == "f1":
#             loss = self.f1_loss(FP, FN)
#         elif self.name == "auc_convex":
#             ordered_fixed = order_fixed_leaves(leaves, splitleaf)
#             if len(ordered_fixed) == 0:
#                 loss = 0
#             else:
#                 Ps, Ns = get_split(leaves, splitleaf)
#                 loss = self.ach_lb(ordered_fixed, Ps, Ns)
#         elif self.name == "partial_auc":
#             ordered_fixed = order_fixed_leaves(leaves, splitleaf)
#             if len(ordered_fixed) == 0:
#                 loss = 1 - theta
#             else:
#                 Ps, Ns = get_split(leaves, splitleaf)
#                 loss = self.pach_lb(ordered_fixed, theta, Ps, Ns)
#         lb = loss + self.lamb * len(leaves)
#         return lb

#     def incre_accu_bound(
#         self,
#         removed_l,
#         new_l1,
#         new_l2,
#         unchanged_leaves=None,
#         removed_leaves=None,
#         w=None,
#         FPu=None,
#         FNu=None,
#         theta=None,
#         ld=None,
#     ):
#         if self.name == "acc":
#             a = self.acc_loss(
#                 removed_l.fp - new_l1.fp - new_l2.fp,
#                 removed_l.fn - new_l1.fn - new_l2.fn,
#             )
#         elif self.name == "bacc":
#             a = self.bacc_loss(
#                 removed_l.fp - new_l1.fp - new_l2.fp,
#                 removed_l.fn - new_l1.fn - new_l2.fn,
#             )
#         elif self.name == "wacc":
#             a = self.wacc_loss(
#                 removed_l.fp - new_l1.fp - new_l2.fp,
#                 removed_l.fn - new_l1.fn - new_l2.fn,
#                 w,
#             )
#         elif self.name == "f1":
#             a = self.f1_loss(FPu + removed_l.fp, FNu + removed_l.fn) - self.f1_loss(
#                 FPu + new_l1.fp + new_l2.fp, FNu + new_l1.fn + new_l2.fn
#             )
#         elif self.name == "auc_convex":
#             removed = removed_leaves.copy()
#             removed.remove(removed_l)
#             _, ld_new = self.ach_loss(unchanged_leaves + removed + [new_l1] + [new_l2])
#             a = ld - ld_new
#         elif self.name == "partial_auc":
#             removed = removed_leaves.copy()
#             removed.remove(removed_l)
#             _, ld_new = self.pach_loss(
#                 unchanged_leaves + removed + [new_l1] + [new_l2], theta
#             )
#             a = ld - ld_new
#         return a

#     def leaf_support_bound(
#         self, leaf, w=None, theta=None, FP=None, FN=None, ld=None, ordered_leaves=None
#     ):
#         if self.name == "acc":
#             tau = self.acc_loss(leaf.fp, leaf.fn)
#         elif self.name == "bacc":
#             tau = self.bacc_loss(leaf.fp, leaf.fn)
#         elif self.name == "wacc":
#             tau = self.wacc_loss(leaf.fp, leaf.fn, w)
#         elif self.name == "f1":
#             tau = ld - self.f1_loss(FP - leaf.fp, FN - leaf.fn)
#         elif self.name == "auc_convex":
#             if leaf.p == 0 or leaf.n == 0:
#                 tau = -1
#             else:
#                 leaves = ordered_leaves.copy()
#                 leaves.remove(leaf)
#                 ld_new = self.ach_lb(leaves, leaf.p, leaf.n)
#                 tau = ld - ld_new
#         elif self.name == "partial_auc":
#             if leaf.p == 0 or leaf.n == 0:
#                 tau = -1
#                 ld_new = ld
#             else:
#                 leaves = ordered_leaves.copy()
#                 leaves.remove(leaf)
#                 ld_new = self.pach_lb(leaves, theta, leaf.p, leaf.n)
#                 tau = ld - ld_new
#         # print('leaf split feature', leaf.rules)
#         # print('tau:',round(tau,4), 'ld:', round(ld,4), 'ld_new:', round(ld_new, 4))
#         return tau

#     def acc_loss(self, FP, FN):
#         return (FP + FN) / (self.P + self.N)

#     def bacc_loss(self, FP, FN):
#         return 0.5 * (FN / self.P + FP / self.N)

#     def wacc_loss(self, FP, FN, w):
#         return (FP + w * FN) / (w * self.P + self.N)

#     def f1_loss(self, FP, FN):
#         return (FP + FN) / (2 * self.P + FP - FN)

#     def ach_loss(self, leaves):
#         ordered_leaves = order_leaves(leaves)
#         tp = fp = np.array([0])
#         if len(leaves) > 1:
#             for i in range(0, len(leaves)):
#                 tp = np.append(tp, tp[i] + ordered_leaves[i].p)
#                 fp = np.append(fp, fp[i] + ordered_leaves[i].n)
#         else:
#             tp = np.append(tp, self.P)
#             fp = np.append(fp, self.N)

#         loss = 1 - 0.5 * sum(
#             [
#                 (tp[i] + tp[i - 1]) * (fp[i] - fp[i - 1]) / (self.P * self.N)
#                 for i in range(1, len(tp))
#             ]
#         )
#         return ordered_leaves, loss

#     def pach_loss(self, leaves, theta):
#         ordered_leaves = order_leaves(leaves)
#         tp = fp = np.array([0], dtype=float)
#         if len(leaves) > 1:
#             i = 0
#             while fp[i] < self.N * theta and i < len(leaves):
#                 tp = np.append(tp, tp[i] + ordered_leaves[i].p)
#                 fp = np.append(fp, fp[i] + ordered_leaves[i].n)
#                 i += 1
#             tp[i] = ((tp[i] - tp[i - 1]) / (fp[i] - fp[i - 1])) * (
#                 self.N * theta - fp[i]
#             ) + tp[i]
#             fp[i] = self.N * theta
#         else:
#             tp = np.append(tp, self.P * theta)
#             fp = np.append(fp, self.N * theta)
#         loss = 1 - 0.5 * sum(
#             [
#                 (tp[i] + tp[i - 1]) * (fp[i] - fp[i - 1]) / (self.P * self.N)
#                 for i in range(1, len(tp))
#             ]
#         )
#         return ordered_leaves, loss

#     def ach_lb(self, leaves, Ps, Ns):
#         tp = np.array([0, Ps], dtype=float)
#         fp = np.array([0, 0], dtype=float)
#         for i in range(len(leaves)):
#             tp = np.append(tp, tp[i + 1] + leaves[i].p)
#             fp = np.append(fp, fp[i + 1] + leaves[i].n)
#         tp = np.append(tp, tp[len(leaves) + 1] + 0)
#         fp = np.append(fp, fp[len(leaves) + 1] + Ns)
#         loss = 1 - 0.5 * sum(
#             [
#                 (tp[i] + tp[i - 1]) * (fp[i] - fp[i - 1]) / (self.P * self.N)
#                 for i in range(1, len(tp))
#             ]
#         )

#         return loss

#     def pach_lb(self, leaves, theta, Ps, Ns):
#         tp = np.array([0, Ps], dtype=float)
#         fp = np.array([0, 0], dtype=float)
#         i = 0
#         while fp[i + 1] < self.N * theta:
#             if i < len(leaves):
#                 tp = np.append(tp, tp[i + 1] + leaves[i].p)
#                 fp = np.append(fp, fp[i + 1] + leaves[i].n)
#                 i += 1
#             else:
#                 tp = np.append(tp, tp[i + 1] + 0)
#                 fp = np.append(fp, fp[i + 1] + Ns)
#                 break
#         tp[len(tp) - 1] = (
#             (tp[len(tp) - 1] - tp[len(tp) - 2]) / (fp[len(fp) - 1] - fp[len(fp) - 2])
#         ) * (self.N * theta - fp[len(fp) - 1]) + tp[len(tp) - 1]
#         fp[len(fp) - 1] = self.N * theta
#         loss = 1 - 0.5 * sum(
#             [
#                 (tp[i] + tp[i - 1]) * (fp[i] - fp[i - 1]) / (self.P * self.N)
#                 for i in range(1, len(tp))
#             ]
#         )
#         return loss


# class CacheTree:
#     def __init__(self, name, P, N, lamb, leaves, w=None, theta=None):
#         self.name = name
#         self.P = P
#         self.N = N
#         self.leaves = leaves
#         self.w = w
#         self.theta = theta

#         bound = Objective(name, P, N, lamb)
#         self.risk = bound.risk(self.leaves, self.w, self.theta)

#     def sorted_leaves(self):
#         return tuple(sorted(leaf.points_cap for leaf in self.leaves))


# class Tree:
#     def __init__(self, cache_tree, n, lamb, splitleaf=None, prior_metric=None):

#         self.cache_tree = cache_tree
#         self.splitleaf = splitleaf
#         leaves = cache_tree.leaves
#         self.H = len(leaves)
#         self.risk = cache_tree.risk

#         bound = Objective(cache_tree.name, cache_tree.P, cache_tree.N, lamb)
#         self.lb = bound.lower_bound(leaves, splitleaf, cache_tree.w, cache_tree.theta)

#         if leaves[0].num_captured == n:
#             self.metric = 0  # null tree
#         elif prior_metric == "objective":
#             self.metric = self.risk
#         elif prior_metric == "bound":
#             self.metric = self.lb
#         elif prior_metric == "curiosity":
#             removed_leaves = list(compress(leaves, splitleaf))  # dsplit
#             num_cap_rm = sum(
#                 leaf.num_captured for leaf in removed_leaves
#             )  # num captured by dsplit
#             if num_cap_rm < n:
#                 self.metric = self.lb / ((n - num_cap_rm) / n)  # supp(dun, xn)
#             else:
#                 self.metric = self.lb / (0.01 / n)  # null tree
#         elif prior_metric == "entropy":
#             removed_leaves = list(compress(leaves, splitleaf))
#             num_cap_rm = sum(leaf.num_captured for leaf in removed_leaves)
#             # entropy weighted by number of points captured
#             self.entropy = [
#                 (
#                     -leaves[i].p * math.log2(leaves[i].p)
#                     - (1 - leaves[i].p) * math.log2(1 - leaves[i].p)
#                 )
#                 * leaves[i].num_captured
#                 if leaves[i].p != 0 and leaves[i].p != 1
#                 else 0
#                 for i in range(self.H)
#             ]
#             if num_cap_rm < n:
#                 self.metric = sum(
#                     self.entropy[i] for i in range(self.H) if splitleaf[i] == 0
#                 ) / (n - sum(leaf.num_captured for leaf in removed_leaves))
#             else:
#                 self.metric = (
#                     sum(self.entropy[i] for i in range(self.H) if splitleaf[i] == 0)
#                     / 0.01
#                 )
#         elif prior_metric == "gini":
#             removed_leaves = list(compress(leaves, splitleaf))
#             num_cap_rm = sum(leaf.num_captured for leaf in removed_leaves)
#             # gini index weighted by number of points captured
#             self.giniindex = [
#                 (2 * leaves[i].p * (1 - leaves[i].p)) * leaves[i].num_captured
#                 for i in range(self.H)
#             ]
#             if num_cap_rm < n:
#                 self.metric = sum(
#                     self.giniindex[i] for i in range(self.H) if splitleaf[i] == 0
#                 ) / (n - sum(leaf.num_captured for leaf in removed_leaves))
#             else:
#                 self.metric = (
#                     sum(self.giniindex[i] for i in range(self.H) if splitleaf[i] == 0)
#                     / 0.01
#                 )
#         elif prior_metric == "FIFO":
#             self.metric = 0
#         elif prior_metric == "random":
#             self.metric = np.random.uniform(0.0, 1.0)

#     def __lt__(self, other):
#         # define <, which will be used in the priority queue
#         return self.metric < other.metric


# class CacheLeaf:
#     def __init__(
#         self,
#         name,
#         n,
#         P,
#         N,
#         rules,
#         x,
#         y,
#         y_mpz,
#         z_mpz,
#         points_cap,
#         num_captured,
#         lamb,
#         support,
#         is_feature_dead,
#         w=None,
#     ):
#         self.rules = rules
#         self.points_cap = points_cap
#         self.num_captured = num_captured
#         self.is_feature_dead = is_feature_dead

#         _, num_ones = rule_vand(points_cap, y_mpz)  # return vand and cnt
#         _, num_errors = rule_vand(points_cap, z_mpz)
#         """
#         print('rules:', rules)
#         print("points_cap:", points_cap, "vec:", rule_mpztovec(points_cap))
#         print('_:', _, "vec:", rule_mpztovec(_))
#         print('num_errors',num_errors)
#         """
#         self.delta = num_errors
#         self.p = num_ones
#         self.n = self.num_captured - num_ones
#         if self.num_captured > 0:
#             self.r = num_ones / self.num_captured
#         else:
#             self.r = 0
#         bound = Objective(name, P, N, lamb)

#         if name == "auc_convex":
#             if num_errors > 0:
#                 cap = np.array(rule_mpztovec(points_cap))
#                 cap_i = np.where(cap == 1)[0]
#                 x_cap = x[cap_i]
#                 y_cap = y[cap_i]

#                 v = rule_mpztovec(_)
#                 equiv_i = np.where(np.array(v) == 1)[0]
#                 idx = [i for i, c in enumerate(cap_i) if c in equiv_i]
#                 idx = np.array(idx)

#                 unique_rows, counts = np.unique(x_cap[idx,], axis=0, return_counts=True)
#                 """
#                 print('cap_i:', cap_i)
#                 print("x_cap:", x_cap)
#                 print("y_cap:", y_cap)
#                 print("v:", v)
#                 print("idx:", idx)
#                 """
#                 nrow = unique_rows.shape[0]
#                 self.equiv = np.zeros((3, nrow + 2))

#                 for i in range(nrow):
#                     comp = np.all(np.equal(x_cap, unique_rows[i,]), axis=1)
#                     eu = np.sum(comp)
#                     j = np.where(comp == True)
#                     n_neg = np.sum(y_cap[j] == 0)
#                     n_pos = eu - n_neg
#                     self.equiv[0, i] = n_pos / eu  # r = n_pos/eu
#                     self.equiv[1, i] = n_pos
#                     self.equiv[2, i] = n_neg

#                 self.equiv[0, nrow] = 1
#                 # y_i = np.where(np.array(v)==0)[0]
#                 # equiv_not_i = [i for i,c in enumerate(cap_i) if c not in equiv_i]
#                 self.equiv[1, nrow] = sum(y_cap == 1) - sum(
#                     self.equiv[1, i] for i in range(nrow)
#                 )
#                 self.equiv[2, nrow + 1] = sum(y_cap == 0) - sum(
#                     self.equiv[2, i] for i in range(nrow)
#                 )
#             else:
#                 self.equiv = np.zeros((3, 2))
#                 self.equiv[0, 0] = 1
#                 self.equiv[1, 0] = self.p
#                 self.equiv[2, 1] = self.n

#         if self.num_captured:
#             self.pred = bound.leaf_predict(self.p, self.n, w)
#             if self.pred == 0:
#                 self.fp = 0
#                 self.fn = self.p
#                 self.delta_fp = 0
#                 self.delta_fn = self.delta
#             else:
#                 self.fp = self.n
#                 self.fn = 0
#                 self.delta_fp = self.delta
#                 self.delta_fn = 0
#         else:
#             self.pred = 0
#             self.fp = 0
#             self.fn = self.p


# class EquivPoints:
#     def __init__(self, r, p, n):
#         self.r = r
#         self.p = p
#         self.n = n


# def get_false(leaves):
#     FP = sum([l.fp for l in leaves])
#     FN = sum([l.fn for l in leaves])
#     return FP, FN


# def get_fixed_false(leaves, splitleaf):
#     FPu = sum([leaves[i].fp for i in range(len(leaves)) if splitleaf[i] == 0])
#     FNu = sum([leaves[i].fn for i in range(len(leaves)) if splitleaf[i] == 0])
#     return FPu, FNu


# def get_split(leaves, splitleaf):
#     Ps = sum([leaves[i].p for i in range(len(leaves)) if splitleaf[i] == 1])
#     Ns = sum([leaves[i].n for i in range(len(leaves)) if splitleaf[i] == 1])
#     return Ps, Ns


# def order_leaves(leaves):
#     return sorted([l for l in leaves], key=lambda x: x.r, reverse=True)


# def order_fixed_leaves(leaves, splitleaf):
#     leaf_fixed = [not split for split in splitleaf]
#     fixed_set = list(compress(leaves, leaf_fixed))
#     ordered_fixed = order_leaves(fixed_set)
#     return ordered_fixed


# def ach_equiv_lb(leaves, splitleaf, P, N, lamb):
#     split_set = list(compress(leaves, splitleaf))
#     leaf_fixed = [not split for split in splitleaf]
#     fixed_set = list(compress(leaves, leaf_fixed))
#     Ps1 = 0
#     Ns1 = 0
#     for i in split_set:
#         equiv = i.equiv  # equiv is the array row=3 and col=equivset+2
#         if equiv.shape[1] - 2 > 0:
#             for j in range(equiv.shape[1] - 2):
#                 leaf_equiv = EquivPoints(equiv[0, j], equiv[1, j], equiv[2, j])
#                 fixed_set.append(leaf_equiv)
#         Ps1 += equiv[1, equiv.shape[1] - 2]
#         Ns1 += equiv[2, equiv.shape[1] - 1]
#     ordered_leaves = order_leaves(fixed_set)

#     tp = np.array([0, Ps1], dtype=float)
#     fp = np.array([0, 0], dtype=float)
#     for i in range(len(ordered_leaves)):
#         tp = np.append(tp, tp[i + 1] + ordered_leaves[i].p)
#         fp = np.append(fp, fp[i + 1] + ordered_leaves[i].n)
#     tp = np.append(tp, tp[len(ordered_leaves) + 1] + 0)
#     fp = np.append(fp, fp[len(ordered_leaves) + 1] + Ns1)

#     loss = 1 - 0.5 * sum(
#         [(tp[i] + tp[i - 1]) * (fp[i] - fp[i - 1]) / (P * N) for i in range(1, len(tp))]
#     )
#     return loss + lamb * len(leaves)


# def generate_new_splitleaf(
#     name,
#     P,
#     N,
#     unchanged_leaves,
#     removed_leaves,
#     new_leaves,
#     lamb,
#     incre_support,
#     w,
#     theta,
# ):

#     n_removed_leaves = len(removed_leaves)  # dsplit
#     n_unchanged_leaves = len(unchanged_leaves)  # dun
#     n_new_leaves = len(new_leaves)
#     n_new_tree_leaves = n_unchanged_leaves + n_new_leaves  # H'
#     splitleaf1 = [0] * n_unchanged_leaves + [1] * n_new_leaves

#     bound = Objective(name, P, N, lamb)
#     FPu = None
#     FNu = None
#     ld = None
#     if name == "f1":
#         FPu, FNu = get_false(unchanged_leaves)
#     if name == "auc_convex":
#         _, ld = bound.ach_loss(unchanged_leaves + removed_leaves)
#     if name == "partial_auc":
#         _, ld = bound.pach_loss(unchanged_leaves + removed_leaves, theta)

#     sl = []
#     for i in range(n_removed_leaves):
#         splitleaf = [0] * n_new_tree_leaves
#         removed_l = removed_leaves[i]
#         new_l1 = new_leaves[2 * i]
#         new_l2 = new_leaves[2 * i + 1]
#         a = bound.incre_accu_bound(
#             removed_l,
#             new_l1,
#             new_l2,
#             unchanged_leaves,
#             removed_leaves,
#             w,
#             FPu,
#             FNu,
#             theta,
#             ld,
#         )
#         if not incre_support:
#             a = float("Inf")
#         if a <= lamb:
#             splitleaf[n_unchanged_leaves + 2 * i] = 1
#             splitleaf[n_unchanged_leaves + 2 * i + 1] = 1
#             sl.append(splitleaf)
#         else:
#             sl.append(splitleaf1)
#     return sl


# def get_cannot_split(name, P, N, lamb, m, new_tree_leaves, MAXDEPTH, w, theta):
#     bound = Objective(name, P, N, lamb)
#     FP = None
#     FN = None
#     ld = None
#     ordered_leaves = None
#     if name == "f1":
#         FP, FN = get_false(new_tree_leaves)
#         ld = bound.f1_loss(FP, FN)
#     if name == "auc_convex":
#         ordered_leaves, ld = bound.ach_loss(new_tree_leaves)
#     if name == "partial_auc":
#         ordered_leaves, ld = bound.pach_loss(new_tree_leaves, theta)

#     cannot_split = [
#         len(l.rules) >= MAXDEPTH
#         or bound.leaf_support_bound(l, w, theta, FP, FN, ld, ordered_leaves) < lamb
#         or all(
#             [
#                 l.is_feature_dead[r - 1]
#                 for r in range(1, m + 1)
#                 if r not in map(abs, l.rules)
#             ]
#         )
#         for l in new_tree_leaves
#     ]
#     return cannot_split


# def bbound(
#     x,
#     y,
#     name,
#     lamb,
#     prior_metric=None,
#     w=None,
#     theta=None,
#     MAXDEPTH=float("Inf"),
#     MAX_NLEAVES=float("Inf"),
#     niter=float("Inf"),
#     logon=False,
#     support=True,
#     incre_support=True,
#     accu_support=True,
#     equiv_points=True,
#     lookahead=True,
#     lenbound=True,
#     R_c0=1,
#     timelimit=float("Inf"),
#     init_cart=True,
#     saveTree=False,
#     readTree=False,
# ):

#     x0 = copy.deepcopy(x)
#     y0 = copy.deepcopy(y)

#     tic = time.time()

#     m = x.shape[1]  # number of features
#     n = len(y)
#     P = np.count_nonzero(y)
#     N = n - P

#     x_mpz = [rule_vectompz(x[:, i]) for i in range(m)]
#     y_mpz = rule_vectompz([i for i in y])

#     # order the columns by descending gini reduction
#     idx, dic = gini_reduction(x_mpz, y_mpz, n, range(m))
#     x = x[:, idx]
#     x_mpz = [x_mpz[i] for i in idx]

#     z_mpz = get_z(x, y, n, m)

#     # lines = []  # a list for log
#     leaf_cache = {}  # cache leaves
#     tree_cache = {}  # cache trees

#     # initialize the queue to include just empty root
#     queue = []
#     root_leaf = CacheLeaf(
#         name,
#         n,
#         P,
#         N,
#         (),
#         x,
#         y,
#         y_mpz,
#         z_mpz,
#         make_all_ones(n + 1),
#         n,
#         lamb,
#         support,
#         [0] * m,
#         w,
#     )
#     d_c = CacheTree(name, P, N, lamb=lamb, leaves=[root_leaf], w=w, theta=theta)
#     R_c = d_c.risk
#     tree0 = Tree(
#         cache_tree=d_c, n=n, lamb=lamb, splitleaf=[1], prior_metric=prior_metric
#     )
#     heapq.heappush(queue, (tree0.metric, tree0))

#     best_is_cart = False  # a flag for whether or not the best is the initial CART
#     if init_cart:
#         clf, nleaves_CART, trainout_CART, R_c, d_c, C_c = cart(
#             x0, y0, name, n, P, N, lamb, w, theta, MAXDEPTH
#         )
#         time_c = time.time() - tic
#         best_is_cart = True
#         # print("risk of cart:", R_c)
#     else:
#         C_c = 0
#         clf = None
#         time_c = time.time()

#     if readTree:
#         with open("tree.pkl", "rb") as f:
#             d_c = pickle.load(f)
#         R_c = d_c.risk

#         with open("leaf_cache.pkl", "rb") as f:
#             leaf_cache = pickle.load(f)

#         sorted_new_tree_rules = tuple(sorted(leaf.points_cap for leaf in d_c.leaves))
#         tree_cache[sorted_new_tree_rules] = True

#         tree_p = Tree(
#             cache_tree=d_c,
#             n=n,
#             lamb=lamb,
#             splitleaf=[1] * len(d_c.leaves),
#             prior_metric=prior_metric,
#         )

#         heapq.heappush(queue, (tree_p.metric, tree_p))
#         """
#         print("PICKEL>>>>>>>>>>>>>", [leaf.rules for leaf in d_c.leaves])
#         print('R_c:', R_c)
#         print('lower_bound:', tree_p.lb)
#         print('lookahead:',tree_p.lb+lamb*sum(tree_p.splitleaf))
#         """
#         # print("leaf_cache:", leaf_cache)

#         C_c = 0
#         time_c = time.time() - tic

#     if R_c0 < R_c:
#         R_c = R_c0

#     leaf_cache[make_all_ones(n + 1)] = root_leaf

#     COUNT = 0  # count the total number of trees in the queue
#     COUNT_POP = 0  # number of tree poped from queue (# of tree checked)
#     COUNT_UNIQLEAVES = 0
#     COUNT_LEAFLOOKUPS = 0

#     bound = Objective(name, P, N, lamb)

#     len_queue = []
#     time_queue = []
#     count_tree = []
#     time_realize_best_tree = [time_c]
#     R_best_tree = [R_c]

#     while queue and COUNT < niter and time.time() - tic < timelimit:
#         """
#         print(len(queue))
#         for metric, t in queue:
#             print(metric, [l.rules for l in t.cache_tree.leaves], t.splitleaf)
#         """
#         metric, tree = heapq.heappop(queue)

#         COUNT_POP = COUNT_POP + 1
#         count_tree.append(COUNT_POP)

#         leaves = tree.cache_tree.leaves
#         leaf_split = tree.splitleaf
#         removed_leaves = list(compress(leaves, leaf_split))
#         old_tree_length = len(leaf_split)
#         new_tree_length = len(leaf_split) + sum(leaf_split)

#         # prefix-specific upper bound on number of leaves
#         if lenbound and new_tree_length >= min(
#             old_tree_length + math.floor((R_c - tree.lb) / lamb) if lamb > 0 else 2**m, 2 ** m
#         ):
#             continue

#         n_removed_leaves = sum(leaf_split)
#         n_unchanged_leaves = old_tree_length - n_removed_leaves

#         # print("num in queue:", len(queue))
#         # print(time.time()-tic)
#         len_queue.append(len(queue))
#         time_queue.append(time.time() - tic)

#         """equivalent points bound + lookahead bound"""
#         lambbb = lamb if lookahead else 0

#         if (name != "auc_convex") & (name != "partial_auc"):
#             delta_fp = (
#                 sum([leaf.delta_fp for leaf in removed_leaves]) if equiv_points else 0
#             )
#             delta_fn = (
#                 sum([leaf.delta_fn for leaf in removed_leaves]) if equiv_points else 0
#             )
#             FPu, FNu = get_fixed_false(leaves, leaf_split)

#             # print("leaf:", [l.rules for l in leaves])
#             # print("leaf fp:", [l.p for l in leaves])
#             # print("leaf fn:", [l.n for l in leaves])
#             # print("leaf delta fp:", [l.delta_fp for l in leaves])
#             # print("leaf delta fn:", [l.delta_fn for l in leaves])
#             # print((delta_fp+delta_fn)/(P+N))
#             # print((FPu+FNu)/(P+N))
#             # print(bound.loss(FPu+delta_fp, FNu+delta_fn, w))
#             # print(n_removed_leaves * lambbb)
#             # print("R_c:", R_c)
#             # print(bound.loss(FPu+delta_fp, FNu+delta_fn, w) + (old_tree_length+n_removed_leaves) * lambbb, R_c)
#             # print(bound.loss(FPu+delta_fp, FPu+delta_fn, w)+ n_removed_leaves * lambbb >= R_c)

#         """
#         if (name != 'auc_convex') and (name != 'partial_auc'):
#             #skip.append(bound.loss(FPu+delta_fp, FNu+delta_fn, w)+ (old_tree_length+n_removed_leaves) * lambbb >= R_c)
#             print(bound.loss(FPu+delta_fp, FNu+delta_fn, w)+ (old_tree_length+n_removed_leaves) * lambbb >= R_c)
#         if (name == 'auc_convex' or name == 'partial_auc'):
#             #skip.append(tree.lb + n_removed_leaves * lambbb>= R_c)
#             print(tree.lb + n_removed_leaves * lambbb>= R_c)
#         """

#         if (
#             (name != "auc_convex")
#             and (name != "partial_auc")
#             and (
#                 bound.loss(FPu + delta_fp, FNu + delta_fn, w)
#                 + (old_tree_length + n_removed_leaves) * lambbb
#                 >= R_c
#             )
#         ):
#             continue

#         if name == "auc_convex":
#             if (
#                 ach_equiv_lb(leaves, leaf_split, P, N, lamb) + n_removed_leaves * lambbb
#                 >= R_c
#             ):
#                 continue

#         if (name == "partial_auc") and (tree.lb + n_removed_leaves * lambbb >= R_c):
#             continue

#         leaf_no_split = [not split for split in leaf_split]
#         unchanged_leaves = list(compress(leaves, leaf_no_split))

#         # Generate all assignments of rules to the leaves that are due to be split
#         rules_for_leaf = [
#             set(range(1, m + 1))
#             - set(map(abs, l.rules))
#             - set([i + 1 for i in range(m) if l.is_feature_dead[i] == 1])
#             for l in removed_leaves
#         ]

#         for leaf_rules in product(*rules_for_leaf):

#             if time.time() - tic >= timelimit:
#                 break

#             new_leaves = []
#             flag_increm = (
#                 False  # a flag for jump out of the loops (incremental support bound)
#             )
#             for rule, removed_leaf in zip(leaf_rules, removed_leaves):

#                 rule_index = rule - 1
#                 tag = (
#                     removed_leaf.points_cap
#                 )  # points captured by the leaf's parent leaf

#                 for new_rule in (-rule, rule):
#                     new_rule_label = int(new_rule > 0)
#                     new_rules = tuple(sorted(removed_leaf.rules + (new_rule,)))
#                     tag_rule = (
#                         x_mpz[rule_index]
#                         if new_rule_label == 1
#                         else ~(x_mpz[rule_index]) | mpz(pow(2, n))
#                     )
#                     new_points_cap, new_num_captured = rule_vand(tag, tag_rule)
#                     leaf_key = new_points_cap
#                     # print(leaf_key)

#                     if leaf_key not in leaf_cache:

#                         COUNT_UNIQLEAVES = COUNT_UNIQLEAVES + 1

#                         # parent_is_feature_dead =
#                         new_leaf = CacheLeaf(
#                             name,
#                             n,
#                             P,
#                             N,
#                             new_rules,
#                             x,
#                             y,
#                             y_mpz,
#                             z_mpz,
#                             new_points_cap,
#                             new_num_captured,
#                             lamb,
#                             support,
#                             removed_leaf.is_feature_dead.copy(),
#                             w,
#                         )
#                         leaf_cache[leaf_key] = new_leaf
#                         new_leaves.append(new_leaf)
#                     else:

#                         COUNT_LEAFLOOKUPS = COUNT_LEAFLOOKUPS + 1

#                         new_leaf = leaf_cache[leaf_key]
#                         new_leaves.append(new_leaf)

#                     """
#                     # Lower bound on classification accuracy
#                     # if (new_leaf.num_captured) / n <= lamb:
#                     # accu_support == theorem 9 in OSDT, check if feature dead, not derived yet
                    
#                     if accu_support == True and (new_leaf.num_captured - new_leaf.num_captured_incorrect) / n <= lamb:

#                         removed_leaf.is_feature_dead[rule_index] = 1

#                         flag_increm = True
#                         break
#                     """

#                 if flag_increm:
#                     break

#             if flag_increm:
#                 continue

#             new_tree_leaves = unchanged_leaves + new_leaves

#             sorted_new_tree_rules = tuple(
#                 sorted(leaf.points_cap for leaf in new_tree_leaves)
#             )

#             if sorted_new_tree_rules in tree_cache:
#                 continue
#             else:
#                 tree_cache[sorted_new_tree_rules] = True

#             child = CacheTree(name, P, N, lamb, new_tree_leaves, w=w, theta=theta)

#             # print([l.rules for l in child.leaves])

#             R = child.risk

#             # print("R:", R, "R_c:", R_c)
#             time_realize_best_tree.append(time.time() - tic)
#             R_best_tree.append(R)

#             if R < R_c:
#                 d_c = child
#                 R_c = R
#                 C_c = COUNT + 1
#                 time_c = time.time() - tic

#                 best_is_cart = False

#             # generate the new splitleaf for the new tree
#             sl = generate_new_splitleaf(
#                 name,
#                 P,
#                 N,
#                 unchanged_leaves,
#                 removed_leaves,
#                 new_leaves,
#                 lamb,
#                 incre_support,
#                 w,
#                 theta,
#             )  # a_j

#             cannot_split = get_cannot_split(
#                 name, P, N, lamb, m, new_tree_leaves, MAXDEPTH, w, theta
#             )

#             # For each copy, we don't split leaves which are not split in its parent tree.
#             # In this way, we can avoid duplications.
#             can_split_leaf = [(0,)] * n_unchanged_leaves + [
#                 (0,) if cannot_split[i] else (0, 1)
#                 for i in range(n_unchanged_leaves, new_tree_length)
#             ]
#             # Discard the first element of leaf_splits, since we must split at least one leaf
#             new_leaf_splits0 = np.array(
#                 list(product(*can_split_leaf))[1:]
#             )  # sorted(product(*can_split_leaf))[1:]
#             len_sl = len(sl)
#             if len_sl == 1:
#                 # Filter out those which split at least one leaf in dp (d0)
#                 new_leaf_splits = [
#                     ls for ls in new_leaf_splits0 if np.dot(ls, sl[0]) > 0
#                 ]
#             else:
#                 # Filter out those which split at least one leaf in dp and split at least one leaf in d0
#                 new_leaf_splits = [
#                     ls
#                     for ls in new_leaf_splits0
#                     if all([np.dot(ls, sl[i]) > 0 for i in range(len_sl)])
#                 ]

#             for new_leaf_split in new_leaf_splits:
#                 # construct the new tree
#                 tree_new = Tree(
#                     cache_tree=child,
#                     n=n,
#                     lamb=lamb,
#                     splitleaf=new_leaf_split,
#                     prior_metric=prior_metric,
#                 )
#                 """
#                 print('tree_lb:', round(tree_new.lb, 4), 
#                       'tree_risk:', round(tree.cache_tree.risk, 4))
#                 """
#                 # print('tree_rules_x8:', [l.rules for l in tree.cache_tree.leaves])

#                 # MAX Number of leaves
#                 if len(new_leaf_split) + sum(new_leaf_split) > MAX_NLEAVES:
#                     continue

#                 COUNT = COUNT + 1
#                 # print([l.rules for l in tree_new.cache_tree.leaves], tree_new.splitleaf)
#                 """
#                 if (COUNT <= 22):
#                     print([l.rules for l in tree_new.cache_tree.leaves], 
#                           tree_new.splitleaf, round(tree_new.lb, 4), 
#                           round(tree_new.cache_tree.risk,4), round(tree_new.metric, 4), 
#                           round(metric,4), [l.rules for l in tree.cache_tree.leaves])
                
#                 if (COUNT ==22)|(COUNT == 21)|(COUNT==20):
#                     for metric, t in queue:
#                         print(metric, [l.rules for l in t.cache_tree.leaves], t.splitleaf)
                   
#                 if COUNT == 22:
#                     print('123455667677')
#                     return
#                 """
#                 # heapq.heappush(queue, (2*tree_new.metric - R_c, tree_new))
#                 heapq.heappush(queue, (tree_new.metric, tree_new))

#                 if COUNT % 1000000 == 0:
#                     # print("COUNT:", COUNT)
#                     pass

#         # print('COUNT:', COUNT)

#     totaltime = time.time() - tic

#     if not best_is_cart:

#         accu = 1 - (R_c - lamb * len(d_c.leaves))

#         leaves_c = [leaf.rules for leaf in d_c.leaves]
#         pred_c = [leaf.pred for leaf in d_c.leaves]

#         num_captured = [leaf.num_captured for leaf in d_c.leaves]

#         # num_captured_incorrect = [leaf.num_captured_incorrect for leaf in d_c.leaves]

#         nleaves = len(leaves_c)
#     else:
#         accu = trainout_CART
#         leaves_c = "NA"
#         pred_c = "NA"
#         get_code(d_c, ["x" + str(i) for i in range(1, m + 1)], [0, 1])
#         num_captured = "NA"
#         # num_captured_incorrect = 'NA'
#         nleaves = nleaves_CART

#     if saveTree:
#         with open("tree.pkl", "wb") as f:
#             pickle.dump(d_c, f)
#         with open("leaf_cache.pkl", "wb") as f:
#             pickle.dump(leaf_cache, f)

#     """
#     print(">>> log:", logon)
#     print(">>> support bound:", support)
#     print(">>> accu_support:", accu_support)
#     print(">>> accurate support bound:", incre_support)
#     print(">>> equiv points bound:", equiv_points)
#     print(">>> lookahead bound:", lookahead)
#     print("prior_metric=", prior_metric)
#     """
#     # print("loss function:", name)
#     # print("lambda: ", lamb)
#     print("COUNT_UNIQLEAVES:", COUNT_UNIQLEAVES)
#     print("COUNT_LEAFLOOKUPS:", COUNT_LEAFLOOKUPS)
#     # print("total time: ", totaltime)
#     # print("leaves: ", leaves_c)
#     # print("num_captured: ", num_captured)
#     # print("prediction: ", pred_c)
#     # print("Objective: ", R_c)
#     # print("Accuracy: ", accu)
#     # print("COUNT of the best tree: ", C_c)
#     # print("time when the best tree is achieved: ", time_c)
#     print("TOTAL COUNT: ", COUNT)

#     return (
#         leaves_c,
#         pred_c,
#         dic,
#         nleaves,
#         m,
#         n,
#         totaltime,
#         time_c,
#         R_c,
#         COUNT,
#         C_c,
#         accu,
#         best_is_cart,
#         clf,
#         len_queue,
#         time_queue,
#         time_realize_best_tree,
#         R_best_tree,
#         count_tree,
#     )


# def predict(
#     name,
#     leaves_c,
#     prediction_c,
#     nleaves,
#     dic,
#     x,
#     y,
#     best_is_cart,
#     clf,
#     w=None,
#     theta=None,
# ):
#     """
#     :param leaves_c:
#     :param dic:
#     :return:
#     """
#     P = np.count_nonzero(y)
#     N = len(y) - P
#     if best_is_cart:
#         yhat = clf.predict(x)

#         n_fp = sum((yhat == 1) & (yhat != y))
#         n_fn = sum((yhat == 0) & (yhat != y))
#         n_tp = sum((yhat == 1) & (yhat == y))

#         if (name != "auc_convex") & (name != "partial_auc"):
#             out = 0.5 * (n_fp * n_tp + (n_tp + P) * (N - n_fp)) / (N * P)
#         elif name == "auc_convex":
#             nodes = clf.apply(x)
#             metric = np.empty([3, int(nleaves)], dtype=float)
#             for i, num in enumerate(set(nodes)):
#                 idx = np.where(nodes == num)
#                 metric[0, i] = sum(y[idx] == 1)
#                 metric[1, i] = sum(y[idx] == 0)
#                 metric[2, i] = sum(y[idx] == 1) / len(y[idx])
#             metric = metric[:, np.argsort(metric[2,])]
#             metric = np.flip(metric, axis=1)
#             metric = np.cumsum(metric, axis=1)
#             init = np.array([[0], [0], [0]])
#             metric = np.append(init, metric, axis=1)
#             tp = metric[0, :]
#             fp = metric[1, :]
#             out = 0.5 * sum(
#                 [
#                     (metric[0, i] / P + metric[0, i - 1] / P)
#                     * (metric[1, i] / N - metric[1, i - 1] / N)
#                     for i in range(1, len(leaves_c) + 1)
#                 ]
#             )
#         elif name == "partial_auc":
#             nodes = clf.apply(x)
#             metric = np.zeros([3, int(nleaves)], dtype=float)
#             for i, num in enumerate(set(nodes)):
#                 idx = np.where(nodes == num)
#                 metric[0, i] = sum(y[idx] == 1)
#                 metric[1, i] = sum(y[idx] == 0)
#                 metric[2, i] = sum(y[idx] == 1) / len(y[idx])
#                 # print(metric[0,i], metric[1,i], metric[2,i])
#             metric = metric[:, np.argsort(metric[2,])]
#             metric = np.flip(metric, axis=1)
#             metric = np.cumsum(metric, axis=1)
#             init = np.array([[0], [0], [0]])
#             metric = np.append(init, metric, axis=1)

#             tp = [0]
#             fp = [0]
#             i = 1
#             while fp[i - 1] < N * theta and i <= int(nleaves) + 1:
#                 tp.append(metric[0, i])
#                 fp.append(metric[1, i])
#                 i += 1

#             tp[i - 1] = ((tp[i - 1] - tp[i - 2]) / (fp[i - 1] - fp[i - 2])) * (
#                 N * theta - fp[i - 1]
#             ) + tp[i - 1]
#             fp[i - 1] = N * theta
#             out = 0.5 * sum(
#                 [
#                     (tp[i] + tp[i - 1]) * (fp[i] - fp[i - 1]) / (P * N)
#                     for i in range(1, len(tp))
#                 ]
#             )

#         print("Best is cart! Testing", "auc:", round(out, 4))
#         print("P=", P, "N=", N, "FP=", n_fp, "FN=", n_fn)
#         if (name == "auc_convex") or (name == "partial_auc"):
#             print("tp=", tp)
#             print("fp=", fp)
#             plt.plot([f / N for f in fp], [f / P for f in tp], "go-", linewidth=2)
#         print(">>>>>>>>>>>>>>>>>>>>>>>")

#         return yhat, out

#     n = x.shape[0]

#     caps = []

#     for leaf in leaves_c:
#         cap = np.array([1] * n)
#         for feature in leaf:
#             idx = dic[abs(feature)]
#             feature_label = int(feature > 0)
#             cap = (x[:, idx] == feature_label) * cap
#         caps.append(cap)

#     yhat = np.array([1] * n)

#     for j in range(len(caps)):
#         idx_cap = [i for i in range(n) if caps[j][i] == 1]
#         yhat[idx_cap] = prediction_c[j]

#     n_fp = sum((yhat == 1) & (yhat != y))
#     n_fn = sum((yhat == 0) & (yhat != y))
#     n_tp = sum((yhat == 1) & (yhat == y))

    
#     # if name == 'acc':
#     #     out = sklearn.metrics.accuracy_score(y, yhat)
#     # elif name == "bacc":
#     #     out = sklearn.metrics.balanced_accuracy_score(y, yhat)
#     # elif name == 'wacc':
#     #     out = 1-(n_fp + w*n_fn)/(len(y))
#     # elif name == 'f1':
#     #     out = sklearn.metrics.f1_score(y, yhat)
    
#     if (name != "auc_convex") & (name != "partial_auc"):
#         out = 0.5 * (n_fp * n_tp + (n_tp + P) * (N - n_fp)) / (N * P)
#     elif name == "auc_convex":
#         n_p = []
#         n_n = []
#         for i in range(len(caps)):
#             idx = np.where(caps[i] == 1)
#             n_p.append(sum(y[idx] == 1))
#             n_n.append(sum(y[idx] == 0))

#         metric = np.empty([3, len(caps)], dtype=float)
#         for i in range(len(caps)):
#             idx = np.where(caps[i] == 1)
#             metric[0, i] = sum(y[idx] == 1)
#             metric[1, i] = sum(y[idx] == 0)
#             metric[2, i] = sum(y[idx] == 1) / len(y[idx])

#         metric = metric[:, np.argsort(metric[2,])]
#         metric = np.flip(metric, axis=1)
#         metric = np.cumsum(metric, axis=1)
#         init = np.array([[0], [0], [0]])
#         metric = np.append(init, metric, axis=1)
#         tp = metric[0, :]
#         fp = metric[1, :]
#         out = 0.5 * sum(
#             [
#                 (metric[0, i] + metric[0, i - 1])
#                 * (metric[1, i] - metric[1, i - 1])
#                 / (P * N)
#                 for i in range(1, len(tp))
#             ]
#         )

#     elif name == "partial_auc":
#         metric = np.empty([3, len(caps)], dtype=float)
#         for i in range(len(caps)):
#             idx = np.where(caps[i] == 1)
#             metric[0, i] = sum(y[idx] == 1)
#             metric[1, i] = sum(y[idx] == 0)
#             metric[2, i] = sum(y[idx] == 1) / len(y[idx])

#         metric = metric[:, np.argsort(metric[2,])]
#         metric = np.flip(metric, axis=1)
#         metric = np.cumsum(metric, axis=1)
#         init = np.array([[0], [0], [0]])
#         metric = np.append(init, metric, axis=1)
#         tp = [0]
#         fp = [0]
#         i = 1
#         while fp[i - 1] < N * theta and i <= len(caps) + 1:
#             tp.append(metric[0, i])
#             fp.append(metric[1, i])
#             i += 1
#         tp[i - 1] = ((tp[i - 1] - tp[i - 2]) / (fp[i - 1] - fp[i - 2])) * (
#             N * theta - fp[i - 1]
#         ) + tp[i - 1]
#         fp[i - 1] = N * theta
#         out = 0.5 * sum(
#             [
#                 (tp[i] / P + tp[i - 1] / P) * (fp[i] / N - fp[i - 1] / N)
#                 for i in range(1, len(tp))
#             ]
#         )

#     print("Testing", "auc:", round(out, 4))
#     if name != "auc_convex" and name != "partial_auc":
#         print("P=", P, "N=", N, "FP=", n_fp, "FN=", n_fn)
#         plt.plot([0, n_fp / N, 1], [0, (P - n_fn) / P, 1], "go-", linewidth=2)
#     else:
#         # print("P=", P, "N=", N, "l.p=", n_p, "l.n=", n_n)
#         print("P=", P, "N=", N, "tp=", tp, "fp=", fp)
#         plt.plot([f / N for f in fp], [f / P for f in tp], "go-", linewidth=2)
#     print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")

#     return yhat, out


